@@ -5,5 +5,7 @@ extern const struct log_info log_info;
 /* Debug Areas of the code */
 enum {
 	DMAIN,
+	DTRXCTRL,
+	DDEV,
 	DLMS,
 };
