@@ -27,6 +27,10 @@
 #include "Transceiver.h"
 #include <Logger.h>
 
+extern "C" {
+#include "osmo_signal.h"
+}
+
 #ifdef HAVE_CONFIG_H
 #include "config.h"
 #endif
@@ -35,12 +39,6 @@ using namespace GSM;
 
 #define USB_LATENCY_INTRVL		10,0
 
-#if USE_UHD
-#  define USB_LATENCY_MIN		6,7
-#else
-#  define USB_LATENCY_MIN		1,1
-#endif
-
 /* Number of running values use in noise average */
 #define NOISE_CNT			20
 
@@ -71,7 +69,7 @@ TransceiverState::~TransceiverState()
   }
 }
 
-bool TransceiverState::init(int filler, size_t sps, float scale, size_t rtsc, unsigned rach_delay)
+bool TransceiverState::init(FillerType filler, size_t sps, float scale, size_t rtsc, unsigned rach_delay)
 {
   signalVector *burst;
 
@@ -81,19 +79,19 @@ bool TransceiverState::init(int filler, size_t sps, float scale, size_t rtsc, un
   for (size_t n = 0; n < 8; n++) {
     for (size_t i = 0; i < 102; i++) {
       switch (filler) {
-      case Transceiver::FILLER_DUMMY:
+      case FILLER_DUMMY:
         burst = generateDummyBurst(sps, n);
         break;
-      case Transceiver::FILLER_NORM_RAND:
+      case FILLER_NORM_RAND:
         burst = genRandNormalBurst(rtsc, sps, n);
         break;
-      case Transceiver::FILLER_EDGE_RAND:
+      case FILLER_EDGE_RAND:
         burst = generateEdgeBurst(rtsc);
         break;
-      case Transceiver::FILLER_ACCESS_RAND:
+      case FILLER_ACCESS_RAND:
         burst = genRandAccessBurst(rach_delay, sps, n);
         break;
-      case Transceiver::FILLER_ZERO:
+      case FILLER_ZERO:
       default:
         burst = generateEmptyBurst(sps, n);
       }
@@ -102,8 +100,8 @@ bool TransceiverState::init(int filler, size_t sps, float scale, size_t rtsc, un
       fillerTable[i][n] = burst;
     }
 
-    if ((filler == Transceiver::FILLER_NORM_RAND) ||
-        (filler == Transceiver::FILLER_EDGE_RAND)) {
+    if ((filler == FILLER_NORM_RAND) ||
+        (filler == FILLER_EDGE_RAND)) {
         chanType[n] = TSC;
     }
   }
@@ -121,7 +119,7 @@ Transceiver::Transceiver(int wBasePort,
   : mBasePort(wBasePort), mLocalAddr(TRXAddress), mRemoteAddr(GSMcoreAddress),
     mClockSocket(TRXAddress, wBasePort, GSMcoreAddress, wBasePort + 100),
     mTransmitLatency(wTransmitLatency), mRadioInterface(wRadioInterface),
-    rssiOffset(wRssiOffset),
+    rssiOffset(wRssiOffset), sig_cbfn(NULL),
     mSPSTx(tx_sps), mSPSRx(rx_sps), mChans(chans), mEdge(false), mOn(false), mForceClockInterface(false),
     mTxFreq(0.0), mRxFreq(0.0), mTSC(0), mMaxExpectedDelayAB(0), mMaxExpectedDelayNB(0),
     mWriteBurstToDiskMask(0)
@@ -161,7 +159,8 @@ Transceiver::~Transceiver()
  * are still expected to report clock indications through control channel
  * activity.
  */
-bool Transceiver::init(int filler, size_t rtsc, unsigned rach_delay, bool edge)
+bool Transceiver::init(FillerType filler, size_t rtsc, unsigned rach_delay,
+                       bool edge, bool ext_rach)
 {
   int d_srcport, d_dstport, c_srcport, c_dstport;
 
@@ -175,6 +174,7 @@ bool Transceiver::init(int filler, size_t rtsc, unsigned rach_delay, bool edge)
     return false;
   }
 
+  mExtRACH = ext_rach;
   mEdge = edge;
 
   mDataSockets.resize(mChans);
@@ -225,6 +225,17 @@ bool Transceiver::init(int filler, size_t rtsc, unsigned rach_delay, bool edge)
   return true;
 }
 
+void Transceiver::setSignalHandler(osmo_signal_cbfn cbfn)
+{
+  if (this->sig_cbfn)
+    osmo_signal_unregister_handler(SS_TRANSC, this->sig_cbfn, NULL);
+
+  if (cbfn) {
+    this->sig_cbfn = cbfn;
+    osmo_signal_register_handler(SS_TRANSC, this->sig_cbfn, NULL);
+  }
+}
+
 /*
  * Start the transceiver
  *
@@ -379,7 +390,8 @@ void Transceiver::pushRadioVector(GSM::Time &nowTime)
     state = &mStates[i];
 
     while ((burst = mTxPriorityQueues[i].getStaleBurst(nowTime))) {
-      LOG(NOTICE) << "dumping STALE burst in TRX->USRP interface";
+      LOGCHAN(i, DMAIN, NOTICE) << "dumping STALE burst in TRX->SDR interface ("
+                  << burst->getTime() <<" vs " << nowTime << "), retrans=" << state->mRetrans;
       if (state->mRetrans)
         updateFillerTable(i, burst);
       delete burst;
@@ -487,16 +499,16 @@ CorrType Transceiver::expectedCorrType(GSM::Time currTime,
     break;
   case IV:
   case VI:
-    return RACH;
+    return mExtRACH ? EXT_RACH : RACH;
     break;
   case V: {
     int mod51 = burstFN % 51;
     if ((mod51 <= 36) && (mod51 >= 14))
-      return RACH;
+      return mExtRACH ? EXT_RACH : RACH;
     else if ((mod51 == 4) || (mod51 == 5))
-      return RACH;
+      return mExtRACH ? EXT_RACH : RACH;
     else if ((mod51 == 45) || (mod51 == 46))
-      return RACH;
+      return mExtRACH ? EXT_RACH : RACH;
     else if (mHandover[burstTN][sdcch4_subslot[burstFN % 102]])
       return RACH;
     else
@@ -514,7 +526,7 @@ CorrType Transceiver::expectedCorrType(GSM::Time currTime,
   case XIII: {
     int mod52 = burstFN % 52;
     if ((mod52 == 12) || (mod52 == 38))
-      return RACH;
+      return mExtRACH ? EXT_RACH : RACH;
     else if ((mod52 == 25) || (mod52 == 51))
       return IDLE;
     else
@@ -625,9 +637,11 @@ SoftVector *Transceiver::pullRadioVector(GSM::Time &wTime, double &RSSI, bool &i
     noise = 20.0 * log10(rxFullScale / state->mNoiseLev);
   }
 
+  unsigned max_toa = (type == RACH || type == EXT_RACH) ?
+                      mMaxExpectedDelayAB : mMaxExpectedDelayNB;
+
   /* Detect normal or RACH bursts */
-  rc = detectAnyBurst(*burst, mTSC, BURST_THRESH, mSPSRx, type, amp, toa,
-                      (type==RACH)?mMaxExpectedDelayAB:mMaxExpectedDelayNB);
+  rc = detectAnyBurst(*burst, mTSC, BURST_THRESH, mSPSRx, type, amp, toa, max_toa);
 
   if (rc > 0) {
     type = (CorrType) rc;
@@ -657,38 +671,66 @@ void Transceiver::reset()
 }
 
 
+#define MAX_PACKET_LENGTH 100
+
+/**
+ * Matches a buffer with a command.
+ * @param  buf    a buffer to look command in
+ * @param  cmd    a command to look in buffer
+ * @param  params pointer to arguments, or NULL
+ * @return        true if command matches, otherwise false
+ */
+static bool match_cmd(char *buf,
+  const char *cmd, char **params)
+{
+  size_t cmd_len = strlen(cmd);
+
+  /* Check a command itself */
+  if (strncmp(buf, cmd, cmd_len))
+    return false;
+
+  /* A command has arguments */
+  if (params != NULL) {
+    /* Make sure there is a space */
+    if (buf[cmd_len] != ' ')
+      return false;
+
+    /* Update external pointer */
+    *params = buf + cmd_len + 1;
+  }
+
+  return true;
+}
+
 void Transceiver::driveControl(size_t chan)
 {
-  int MAX_PACKET_LENGTH = 100;
+  char buffer[MAX_PACKET_LENGTH + 1];
+  char response[MAX_PACKET_LENGTH + 1];
+  char *command, *params;
+  int msgLen;
 
-  // check control socket
-  char buffer[MAX_PACKET_LENGTH];
-  int msgLen = -1;
-  buffer[0] = '\0';
+  /* Attempt to read from control socket */
+  msgLen = mCtrlSockets[chan]->read(buffer, MAX_PACKET_LENGTH);
+  if (msgLen < 1)
+    return;
 
-  msgLen = mCtrlSockets[chan]->read(buffer, sizeof(buffer));
+  /* Zero-terminate received string */
+  buffer[msgLen] = '\0';
 
-  if (msgLen < 1) {
+  /* Verify a command signature */
+  if (strncmp(buffer, "CMD ", 4)) {
+    LOGC(DTRXCTRL, WARNING) << "bogus message on control interface";
     return;
   }
 
-  char cmdcheck[4];
-  char command[MAX_PACKET_LENGTH];
-  char response[MAX_PACKET_LENGTH];
+  /* Set command pointer */
+  command = buffer + 4;
+  LOGCHAN(chan, DTRXCTRL, INFO) << "command is '" << command << "'";
 
-  sscanf(buffer,"%3s %s",cmdcheck,command);
-
-  if (strcmp(cmdcheck,"CMD")!=0) {
-    LOG(WARNING) << "bogus message on control interface";
-    return;
-  }
-  LOG(INFO) << "command is " << buffer;
-
-  if (strcmp(command,"POWEROFF")==0) {
+  if (match_cmd(command, "POWEROFF", NULL)) {
     stop();
     sprintf(response,"RSP POWEROFF 0");
-  }
-  else if (strcmp(command,"POWERON")==0) {
+  } else if (match_cmd(command, "POWERON", NULL)) {
     if (!start()) {
       sprintf(response,"RSP POWERON 1");
     } else {
@@ -698,41 +740,43 @@ void Transceiver::driveControl(size_t chan)
           mHandover[i][j] = false;
       }
     }
-  }
-  else if (strcmp(command,"HANDOVER")==0){
-    int ts=0,ss=0;
-    sscanf(buffer,"%3s %s %d %d",cmdcheck,command,&ts,&ss);
-    mHandover[ts][ss] = true;
-    sprintf(response,"RSP HANDOVER 0 %d %d",ts,ss);
-  }
-  else if (strcmp(command,"NOHANDOVER")==0){
-    int ts=0,ss=0;
-    sscanf(buffer,"%3s %s %d %d",cmdcheck,command,&ts,&ss);
-    mHandover[ts][ss] = false;
-    sprintf(response,"RSP NOHANDOVER 0 %d %d",ts,ss);
-  }
-  else if (strcmp(command,"SETMAXDLY")==0) {
+  } else if (match_cmd(command, "HANDOVER", &params)) {
+    unsigned ts = 0, ss = 0;
+    sscanf(params, "%u %u", &ts, &ss);
+    if (ts > 7 || ss > 7) {
+      sprintf(response, "RSP NOHANDOVER 1 %u %u", ts, ss);
+    } else {
+      mHandover[ts][ss] = true;
+      sprintf(response, "RSP HANDOVER 0 %u %u", ts, ss);
+    }
+  } else if (match_cmd(command, "NOHANDOVER", &params)) {
+    unsigned ts = 0, ss = 0;
+    sscanf(params, "%u %u", &ts, &ss);
+    if (ts > 7 || ss > 7) {
+      sprintf(response, "RSP NOHANDOVER 1 %u %u", ts, ss);
+    } else {
+      mHandover[ts][ss] = false;
+      sprintf(response, "RSP NOHANDOVER 0 %u %u", ts, ss);
+    }
+  } else if (match_cmd(command, "SETMAXDLY", &params)) {
     //set expected maximum time-of-arrival
     int maxDelay;
-    sscanf(buffer,"%3s %s %d",cmdcheck,command,&maxDelay);
+    sscanf(params, "%d", &maxDelay);
     mMaxExpectedDelayAB = maxDelay; // 1 GSM symbol is approx. 1 km
     sprintf(response,"RSP SETMAXDLY 0 %d",maxDelay);
-  }
-  else if (strcmp(command,"SETMAXDLYNB")==0) {
+  } else if (match_cmd(command, "SETMAXDLYNB", &params)) {
     //set expected maximum time-of-arrival
     int maxDelay;
-    sscanf(buffer,"%3s %s %d",cmdcheck,command,&maxDelay);
+    sscanf(params, "%d", &maxDelay);
     mMaxExpectedDelayNB = maxDelay; // 1 GSM symbol is approx. 1 km
     sprintf(response,"RSP SETMAXDLYNB 0 %d",maxDelay);
-  }
-  else if (strcmp(command,"SETRXGAIN")==0) {
+  } else if (match_cmd(command, "SETRXGAIN", &params)) {
     //set expected maximum time-of-arrival
     int newGain;
-    sscanf(buffer,"%3s %s %d",cmdcheck,command,&newGain);
+    sscanf(params, "%d", &newGain);
     newGain = mRadioInterface->setRxGain(newGain, chan);
     sprintf(response,"RSP SETRXGAIN 0 %d",newGain);
-  }
-  else if (strcmp(command,"NOISELEV")==0) {
+  } else if (match_cmd(command, "NOISELEV", NULL)) {
     if (mOn) {
       float lev = mStates[chan].mNoiseLev;
       sprintf(response,"RSP NOISELEV 0 %d",
@@ -741,86 +785,78 @@ void Transceiver::driveControl(size_t chan)
     else {
       sprintf(response,"RSP NOISELEV 1  0");
     }
-  }
-  else if (!strcmp(command, "SETPOWER")) {
+  } else if (match_cmd(command, "SETPOWER", &params)) {
     int power;
-    sscanf(buffer, "%3s %s %d", cmdcheck, command, &power);
+    sscanf(params, "%d", &power);
     power = mRadioInterface->setPowerAttenuation(power, chan);
     mStates[chan].mPower = power;
     sprintf(response, "RSP SETPOWER 0 %d", power);
-  }
-  else if (!strcmp(command,"ADJPOWER")) {
+  } else if (match_cmd(command, "ADJPOWER", &params)) {
     int power, step;
-    sscanf(buffer, "%3s %s %d", cmdcheck, command, &step);
+    sscanf(params, "%d", &step);
     power = mStates[chan].mPower + step;
     power = mRadioInterface->setPowerAttenuation(power, chan);
     mStates[chan].mPower = power;
     sprintf(response, "RSP ADJPOWER 0 %d", power);
-  }
-  else if (strcmp(command,"RXTUNE")==0) {
+  } else if (match_cmd(command, "RXTUNE", &params)) {
     // tune receiver
     int freqKhz;
-    sscanf(buffer,"%3s %s %d",cmdcheck,command,&freqKhz);
+    sscanf(params, "%d", &freqKhz);
     mRxFreq = freqKhz * 1e3;
     if (!mRadioInterface->tuneRx(mRxFreq, chan)) {
-       LOG(ALERT) << "RX failed to tune";
+       LOGC(DTRXCTRL, ALERT) << "RX failed to tune";
        sprintf(response,"RSP RXTUNE 1 %d",freqKhz);
     }
     else
        sprintf(response,"RSP RXTUNE 0 %d",freqKhz);
-  }
-  else if (strcmp(command,"TXTUNE")==0) {
+  } else if (match_cmd(command, "TXTUNE", &params)) {
     // tune txmtr
     int freqKhz;
-    sscanf(buffer,"%3s %s %d",cmdcheck,command,&freqKhz);
+    sscanf(params, "%d", &freqKhz);
     mTxFreq = freqKhz * 1e3;
     if (!mRadioInterface->tuneTx(mTxFreq, chan)) {
-       LOG(ALERT) << "TX failed to tune";
+       LOGC(DTRXCTRL, ALERT) << "TX failed to tune";
        sprintf(response,"RSP TXTUNE 1 %d",freqKhz);
     }
     else
        sprintf(response,"RSP TXTUNE 0 %d",freqKhz);
-  }
-  else if (!strcmp(command,"SETTSC")) {
+  } else if (match_cmd(command, "SETTSC", &params)) {
     // set TSC
     unsigned TSC;
-    sscanf(buffer, "%3s %s %d", cmdcheck, command, &TSC);
+    sscanf(params, "%u", &TSC);
     if (TSC > 7) {
       sprintf(response, "RSP SETTSC 1 %d", TSC);
     } else {
-      LOG(NOTICE) << "Changing TSC from " << mTSC << " to " << TSC;
+      LOGC(DTRXCTRL, NOTICE) << "Changing TSC from " << mTSC << " to " << TSC;
       mTSC = TSC;
       sprintf(response,"RSP SETTSC 0 %d", TSC);
     }
-  }
-  else if (strcmp(command,"SETSLOT")==0) {
+  } else if (match_cmd(command, "SETSLOT", &params)) {
     // set slot type
     int  corrCode;
     int  timeslot;
-    sscanf(buffer,"%3s %s %d %d",cmdcheck,command,&timeslot,&corrCode);
+    sscanf(params, "%d %d", &timeslot, &corrCode);
     if ((timeslot < 0) || (timeslot > 7)) {
-      LOG(WARNING) << "bogus message on control interface";
+      LOGC(DTRXCTRL, WARNING) << "bogus message on control interface";
       sprintf(response,"RSP SETSLOT 1 %d %d",timeslot,corrCode);
       return;
     }
     mStates[chan].chanType[timeslot] = (ChannelCombination) corrCode;
     setModulus(timeslot, chan);
     sprintf(response,"RSP SETSLOT 0 %d %d",timeslot,corrCode);
-
-  }
-  else if (strcmp(command,"_SETBURSTTODISKMASK")==0) {
+  } else if (match_cmd(command, "_SETBURSTTODISKMASK", &params)) {
     // debug command! may change or disapear without notice
     // set a mask which bursts to dump to disk
     int mask;
-    sscanf(buffer,"%3s %s %d",cmdcheck,command,&mask);
+    sscanf(params, "%d", &mask);
     mWriteBurstToDiskMask = mask;
     sprintf(response,"RSP _SETBURSTTODISKMASK 0 %d",mask);
-  }
-  else {
-    LOG(WARNING) << "bogus command " << command << " on control interface.";
+  } else {
+    LOGC(DTRXCTRL, WARNING) << "bogus command " << command << " on control interface.";
     sprintf(response,"RSP ERR 1");
   }
 
+  LOGCHAN(chan, DTRXCTRL, INFO) << "response is '" << response << "'";
   mCtrlSockets[chan]->write(response, strlen(response) + 1);
 }
 
@@ -869,8 +905,12 @@ bool Transceiver::driveTxPriorityQueue(size_t chan)
 
 void Transceiver::driveReceiveRadio()
 {
-  if (!mRadioInterface->driveReceiveRadio()) {
+  int rc = mRadioInterface->driveReceiveRadio();
+  if (rc == 0) {
     usleep(100000);
+  } else if (rc < 0) {
+    LOG(FATAL) << "radio Interface receive failed, requesting stop.";
+    osmo_signal_dispatch(SS_TRANSC, S_TRANSC_STOP_REQUIRED, this);
   } else if (mForceClockInterface || mTransmitDeadlineClock > mLastClockUpdateTime + GSM::Time(216,0)) {
     mForceClockInterface = false;
     writeClockInterface();
@@ -966,14 +1006,15 @@ void Transceiver::driveTxFIFO()
           // only update latency at the defined frame interval
           if (radioClock->get() > mLatencyUpdateTime + GSM::Time(USB_LATENCY_INTRVL)) {
             mTransmitLatency = mTransmitLatency + GSM::Time(1,0);
-            LOG(INFO) << "new latency: " << mTransmitLatency;
+            LOG(INFO) << "new latency: " << mTransmitLatency << " (underrun "
+                      << radioClock->get() << " vs " << mLatencyUpdateTime + GSM::Time(USB_LATENCY_INTRVL) << ")";
             mLatencyUpdateTime = radioClock->get();
           }
         }
         else {
           // if underrun hasn't occurred in the last sec (216 frames) drop
           //    transmit latency by a timeslot
-          if (mTransmitLatency > GSM::Time(USB_LATENCY_MIN)) {
+          if (mTransmitLatency > mRadioInterface->minLatency()) {
               if (radioClock->get() > mLatencyUpdateTime + GSM::Time(216,0)) {
               mTransmitLatency.decTN();
               LOG(INFO) << "reduced latency: " << mTransmitLatency;
@@ -1009,11 +1050,15 @@ void Transceiver::writeClockInterface()
 
 void *RxUpperLoopAdapter(TransceiverChannel *chan)
 {
+  char thread_name[16];
   Transceiver *trx = chan->trx;
   size_t num = chan->num;
 
   delete chan;
 
+  snprintf(thread_name, 16, "RxUpper%zu", num);
+  set_selfthread_name(thread_name);
+
   trx->setPriority(0.42);
 
   while (1) {
@@ -1025,6 +1070,8 @@ void *RxUpperLoopAdapter(TransceiverChannel *chan)
 
 void *RxLowerLoopAdapter(Transceiver *transceiver)
 {
+  set_selfthread_name("RxLower");
+
   transceiver->setPriority(0.45);
 
   while (1) {
@@ -1036,6 +1083,8 @@ void *RxLowerLoopAdapter(Transceiver *transceiver)
 
 void *TxLowerLoopAdapter(Transceiver *transceiver)
 {
+  set_selfthread_name("TxLower");
+
   transceiver->setPriority(0.44);
 
   while (1) {
@@ -1047,11 +1096,15 @@ void *TxLowerLoopAdapter(Transceiver *transceiver)
 
 void *ControlServiceLoopAdapter(TransceiverChannel *chan)
 {
+  char thread_name[16];
   Transceiver *trx = chan->trx;
   size_t num = chan->num;
 
   delete chan;
 
+  snprintf(thread_name, 16, "CtrlService%zu", num);
+  set_selfthread_name(thread_name);
+
   while (1) {
     trx->driveControl(num);
     pthread_testcancel();
@@ -1061,11 +1114,15 @@ void *ControlServiceLoopAdapter(TransceiverChannel *chan)
 
 void *TxUpperLoopAdapter(TransceiverChannel *chan)
 {
+  char thread_name[16];
   Transceiver *trx = chan->trx;
   size_t num = chan->num;
 
   delete chan;
 
+  snprintf(thread_name, 16, "TxUpper%zu", num);
+  set_selfthread_name(thread_name);
+
   trx->setPriority(0.40);
 
   while (1) {
