@@ -21,6 +21,7 @@
 #include "Logger.h"
 #include "Threads.h"
 #include "LMSDevice.h"
+#include "Utils.h"
 
 #include <lime/LimeSuite.h>
 
@@ -39,6 +40,7 @@ constexpr double LMSDevice::masterClockRate;
 #define GSM_CARRIER_BW 270000.0 /* 270kHz */
 #define LMS_MIN_BW_SUPPORTED 2.5e6 /* 2.5mHz, minimum supported by LMS */
 #define LMS_CALIBRATE_BW_HZ OSMO_MAX(GSM_CARRIER_BW, LMS_MIN_BW_SUPPORTED)
+#define SAMPLE_BUF_SZ    (1 << 20) /* Size of Rx timestamp based Ring buffer, in bytes */
 
 LMSDevice::LMSDevice(size_t tx_sps, size_t rx_sps, InterfaceType iface, size_t chans, double lo_offset,
 		     const std::vector<std::string>& tx_paths,
@@ -46,15 +48,35 @@ LMSDevice::LMSDevice(size_t tx_sps, size_t rx_sps, InterfaceType iface, size_t c
 	RadioDevice(tx_sps, rx_sps, iface, chans, lo_offset, tx_paths, rx_paths),
 	m_lms_dev(NULL)
 {
-	LOG(INFO) << "creating LMS device...";
+	LOGC(DDEV, INFO) << "creating LMS device...";
 
 	m_lms_stream_rx.resize(chans);
 	m_lms_stream_tx.resize(chans);
 
 	m_last_rx_underruns.resize(chans, 0);
 	m_last_rx_overruns.resize(chans, 0);
+	m_last_rx_dropped.resize(chans, 0);
 	m_last_tx_underruns.resize(chans, 0);
-	m_last_tx_overruns.resize(chans, 0);
+
+	rx_buffers.resize(chans);
+}
+
+LMSDevice::~LMSDevice()
+{
+	unsigned int i;
+	LOGC(DDEV, INFO) << "Closing LMS device";
+	if (m_lms_dev) {
+		/* disable all channels */
+		for (i=0; i<chans; i++) {
+			LMS_EnableChannel(m_lms_dev, LMS_CH_RX, i, false);
+			LMS_EnableChannel(m_lms_dev, LMS_CH_TX, i, false);
+		}
+		LMS_Close(m_lms_dev);
+		m_lms_dev = NULL;
+	}
+
+	for (size_t i = 0; i < rx_buffers.size(); i++)
+		delete rx_buffers[i];
 }
 
 static void lms_log_callback(int lvl, const char *msg)
@@ -62,169 +84,208 @@ static void lms_log_callback(int lvl, const char *msg)
 	/* map lime specific log levels */
 	static const int lvl_map[5] = {
 		[0] = LOGL_FATAL,
-		[1] = LOGL_ERROR,
-		[2] = LOGL_NOTICE,
-		[3] = LOGL_INFO,
-		[4] = LOGL_DEBUG,
+		[LMS_LOG_ERROR] = LOGL_ERROR,
+		[LMS_LOG_WARNING] = LOGL_NOTICE,
+		[LMS_LOG_INFO] = LOGL_INFO,
+		[LMS_LOG_DEBUG] = LOGL_DEBUG,
 	};
 	/* protect against future higher log level values (lower importance) */
 	if ((unsigned int) lvl >= ARRAY_SIZE(lvl_map))
 		lvl = ARRAY_SIZE(lvl_map)-1;
 
-	LOGLV(DLMS, lvl) << msg;
-}
-
-static void thread_enable_cancel(bool cancel)
-{
-	cancel ? pthread_setcancelstate(PTHREAD_CANCEL_ENABLE, NULL) :
-		 pthread_setcancelstate(PTHREAD_CANCEL_DISABLE, NULL);
+	LOGLV(DLMS, lvl_map[lvl]) << msg;
 }
 
 static void print_range(const char* name, lms_range_t *range)
 {
-	LOG(DEBUG) << name << ": Min=" << range->min << " Max=" << range->max
+	LOGC(DDEV, INFO) << name << ": Min=" << range->min << " Max=" << range->max
 		   << " Step=" << range->step;
 }
 
+/*! Find the device string that matches all filters from \a args.
+ *  \param[in] info_list device addresses found by LMS_GetDeviceList()
+ *  \param[in] count length of info_list
+ *  \param[in] args dev-args value from osmo-trx.cfg, containing comma separated key=value pairs
+ *  \return index of first matching device or -1 (no match) */
+int info_list_find(lms_info_str_t* info_list, unsigned int count, const std::string &args)
+{
+	unsigned int i, j;
+	vector<string> filters;
+
+	filters = comma_delimited_to_vector(args.c_str());
+
+	/* iterate over device addresses */
+	for (i=0; i < count; i++) {
+		/* check if all filters match */
+		bool match = true;
+		for (j=0; j < filters.size(); j++) {
+			if (!strstr(info_list[i], filters[j].c_str())) {
+				match = false;
+				break;
+			}
+		}
+
+		if (match)
+			return i;
+	}
+	return -1;
+}
+
 int LMSDevice::open(const std::string &args, int ref, bool swap_channels)
 {
-	//lms_info_str_t dev_str;
 	lms_info_str_t* info_list;
-	lms_range_t range_lpfbw_rx, range_lpfbw_tx, range_sr;
-	float_type sr_host, sr_rf, lpfbw_rx, lpfbw_tx;
-	uint16_t dac_val;
+	const lms_dev_info_t* device_info;
+	lms_range_t range_sr;
+	float_type sr_host, sr_rf;
 	unsigned int i, n;
-	int rc;
+	int rc, dev_id;
 
-	LOG(INFO) << "Opening LMS device..";
+	LOGC(DDEV, INFO) << "Opening LMS device..";
 
 	LMS_RegisterLogHandler(&lms_log_callback);
 
 	if ((n = LMS_GetDeviceList(NULL)) < 0)
-		LOG(ERROR) << "LMS_GetDeviceList(NULL) failed";
-	LOG(DEBUG) << "Devices found: " << n;
+		LOGC(DDEV, ERROR) << "LMS_GetDeviceList(NULL) failed";
+	LOGC(DDEV, INFO) << "Devices found: " << n;
 	if (n < 1)
 	    return -1;
 
 	info_list = new lms_info_str_t[n];
 
 	if (LMS_GetDeviceList(info_list) < 0)
-		LOG(ERROR) << "LMS_GetDeviceList(info_list) failed";
+		LOGC(DDEV, ERROR) << "LMS_GetDeviceList(info_list) failed";
 
 	for (i = 0; i < n; i++)
-		LOG(DEBUG) << "Device [" << i << "]: " << info_list[i];
+		LOGC(DDEV, INFO) << "Device [" << i << "]: " << info_list[i];
 
-	rc = LMS_Open(&m_lms_dev, info_list[0], NULL);
+	dev_id = info_list_find(info_list, n, args);
+	if (dev_id == -1) {
+		LOGC(DDEV, ERROR) << "No LMS device found with address '" << args << "'";
+		delete[] info_list;
+		return -1;
+	}
+
+	LOGC(DDEV, INFO) << "Using device[" << dev_id << "]";
+	rc = LMS_Open(&m_lms_dev, info_list[dev_id], NULL);
 	if (rc != 0) {
-		LOG(ERROR) << "LMS_GetDeviceList() failed)";
+		LOGC(DDEV, ERROR) << "LMS_GetDeviceList() failed)";
 		delete [] info_list;
 		return -1;
 	}
 
 	delete [] info_list;
 
-	LOG(INFO) << "Init LMS device";
-	if (LMS_Init(m_lms_dev) != 0) {
-		LOG(ERROR) << "LMS_Init() failed";
-		return -1;
+	device_info = LMS_GetDeviceInfo(m_lms_dev);
+
+	if ((ref != REF_EXTERNAL) && (ref != REF_INTERNAL)){
+		LOGC(DDEV, ERROR) << "Invalid reference type";
+		goto out_close;
 	}
 
+	/* if reference clock is external setup must happen _before_ calling LMS_Init */
+	/* FIXME make external reference frequency configurable */
+	if (ref == REF_EXTERNAL) {
+		LOGC(DDEV, INFO) << "Setting External clock reference to 10MHz";
+		/* Assume an external 10 MHz reference clock */
+		if (LMS_SetClockFreq(m_lms_dev, LMS_CLOCK_EXTREF, 10000000.0) < 0)
+			goto out_close;
+	}
+
+	LOGC(DDEV, INFO) << "Init LMS device";
+	if (LMS_Init(m_lms_dev) != 0) {
+		LOGC(DDEV, ERROR) << "LMS_Init() failed";
+		goto out_close;
+	}
+
+	/* LimeSDR-Mini does not have switches but needs soldering to select external/internal clock */
+	/* LimeNET-Micro also does not like selecting internal clock*/
+	/* also set device specific maximum tx levels selected by phasenoise measurements*/
+	if (strncmp(device_info->deviceName,"LimeSDR-USB",11) == 0){
+		/* if reference clock is internal setup must happen _after_ calling LMS_Init */
+		/* according to lms using LMS_CLOCK_EXTREF with a frequency <= 0 is the correct way to set clock to internal reference*/
+		if (ref == REF_INTERNAL) {
+			LOGC(DDEV, INFO) << "Setting Internal clock reference";
+			if (LMS_SetClockFreq(m_lms_dev, LMS_CLOCK_EXTREF, -1) < 0)
+				goto out_close;
+		}
+		maxTxGainClamp = 73.0;
+	} else if (strncmp(device_info->deviceName,"LimeSDR-Mini",12) == 0)
+		maxTxGainClamp = 66.0;
+	else
+		maxTxGainClamp = 71.0; /* "LimeNET-Micro", etc FIXME pciE based LMS boards?*/
+
+	/* enable all used channels */
+	for (i=0; i<chans; i++) {
+		if (LMS_EnableChannel(m_lms_dev, LMS_CH_RX, i, true) < 0)
+			goto out_close;
+		if (LMS_EnableChannel(m_lms_dev, LMS_CH_TX, i, true) < 0)
+			goto out_close;
+	}
+
+	/* set samplerate */
 	if (LMS_GetSampleRateRange(m_lms_dev, LMS_CH_RX, &range_sr))
 		goto out_close;
 	print_range("Sample Rate", &range_sr);
 
-	LOG(DEBUG) << "Setting sample rate to " << GSMRATE*tx_sps << " " << tx_sps;
+	LOGC(DDEV, INFO) << "Setting sample rate to " << GSMRATE*tx_sps << " " << tx_sps;
 	if (LMS_SetSampleRate(m_lms_dev, GSMRATE*tx_sps, 32) < 0)
 		goto out_close;
 
 	if (LMS_GetSampleRate(m_lms_dev, LMS_CH_RX, 0, &sr_host, &sr_rf))
 		goto out_close;
-	LOG(DEBUG) << "Sample Rate: Host=" << sr_host << " RF=" << sr_rf;
+	LOGC(DDEV, INFO) << "Sample Rate: Host=" << sr_host << " RF=" << sr_rf;
 
 	/* FIXME: make this device/model dependent, like UHDDevice:dev_param_map! */
 	ts_offset = static_cast<TIMESTAMP>(8.9e-5 * GSMRATE * tx_sps); /* time * sample_rate */
 
-	switch (ref) {
-	case REF_INTERNAL:
-		LOG(DEBUG) << "Setting Internal clock reference";
-		/* Ugly API: Selecting clock source implicit by writing to VCTCXO DAC ?!? */
-		if (LMS_VCTCXORead(m_lms_dev, &dac_val) < 0)
-			goto out_close;
-		LOG(DEBUG) << "Setting VCTCXO to " << dac_val;
-		if (LMS_VCTCXOWrite(m_lms_dev, dac_val) < 0)
-			goto out_close;
-		break;
-	case REF_EXTERNAL:
-		LOG(DEBUG) << "Setting External clock reference to " << 10000000.0;
-		/* Assume an external 10 MHz reference clock */
-		if (LMS_SetClockFreq(m_lms_dev, LMS_CLOCK_EXTREF, 10000000.0) < 0)
-			goto out_close;
-		break;
-	default:
-		LOG(ALERT) << "Invalid reference type";
-		goto out_close;
-	}
-
-	if (LMS_GetLPFBWRange(m_lms_dev, LMS_CH_RX, &range_lpfbw_rx))
-		goto out_close;
-	print_range("LPFBWRange Rx", &range_lpfbw_rx);
-	if (LMS_GetLPFBWRange(m_lms_dev, LMS_CH_RX, &range_lpfbw_tx))
-		goto out_close;
-	print_range("LPFBWRange Tx", &range_lpfbw_tx);
-	lpfbw_rx = OSMO_MIN(OSMO_MAX(1.4001e6, range_lpfbw_rx.min), range_lpfbw_rx.max);
-	lpfbw_tx = OSMO_MIN(OSMO_MAX(5.2e6, range_lpfbw_tx.min), range_lpfbw_tx.max);
-
-	LOG(DEBUG) << "LPFBW: Rx=" << lpfbw_rx << " Tx=" << lpfbw_tx;
-
+	/* configure antennas */
 	if (!set_antennas()) {
-		LOG(ALERT) << "LMS antenna setting failed";
-		return -1;
+		LOGC(DDEV, FATAL) << "LMS antenna setting failed";
+		goto out_close;
 	}
 
-	/* Perform Rx and Tx calibration */
-	for (i=0; i<chans; i++) {
-		LOG(INFO) << "Setting LPFBW chan " << i;
-		if (LMS_SetLPFBW(m_lms_dev, LMS_CH_RX, i, lpfbw_rx) < 0)
-			goto out_close;
-		if (LMS_SetLPFBW(m_lms_dev, LMS_CH_TX, i, lpfbw_tx) < 0)
-			goto out_close;
-		LOG(INFO) << "Calibrating chan " << i;
-		if (LMS_Calibrate(m_lms_dev, LMS_CH_RX, i, LMS_CALIBRATE_BW_HZ, 0) < 0)
-			goto out_close;
-		if (LMS_Calibrate(m_lms_dev, LMS_CH_TX, i, LMS_CALIBRATE_BW_HZ, 0) < 0)
-			goto out_close;
-	}
+	/* Set up per-channel Rx timestamp based Ring buffers */
+	for (size_t i = 0; i < rx_buffers.size(); i++)
+		rx_buffers[i] = new smpl_buf(SAMPLE_BUF_SZ / sizeof(uint32_t));
 
-	samplesRead = 0;
-	samplesWritten = 0;
 	started = false;
 
 	return NORMAL;
 
 out_close:
-	LOG(ALERT) << "Error in LMS open, closing: " << LMS_GetLastErrorMessage();
+	LOGC(DDEV, FATAL) << "Error in LMS open, closing: " << LMS_GetLastErrorMessage();
 	LMS_Close(m_lms_dev);
+	m_lms_dev = NULL;
 	return -1;
 }
 
 bool LMSDevice::start()
 {
-	LOG(INFO) << "starting LMS...";
+	LOGC(DDEV, INFO) << "starting LMS...";
 
 	unsigned int i;
 
+	if (started) {
+		LOGC(DDEV, ERR) << "Device already started";
+		return false;
+	}
+
 	/* configure the channels/streams */
 	for (i=0; i<chans; i++) {
-		if (LMS_EnableChannel(m_lms_dev, LMS_CH_RX, i, true) < 0)
+		/* Set gains for calibration/filter setup */
+		/* TX gain to maximum */
+		setTxGain(maxTxGain(), i);
+		/* RX gain to midpoint */
+		setRxGain((minRxGain() + maxRxGain()) / 2, i);
+
+		/* set up Rx and Tx filters */
+		if (!do_filters(i))
+			return false;
+		/* Perform Rx and Tx calibration */
+		if (!do_calib(i))
 			return false;
 
-		if (LMS_EnableChannel(m_lms_dev, LMS_CH_TX, i, true) < 0)
-			return false;
-
-		// Set gains to midpoint
-		setTxGain((minTxGain() + maxTxGain()) / 2, i);
-		setRxGain(34.0, i);
-
+		/* configure Streams */
 		m_lms_stream_rx[i] = {};
 		m_lms_stream_rx[i].isTx = false;
 		m_lms_stream_rx[i].channel = i;
@@ -272,17 +333,59 @@ bool LMSDevice::stop()
 	for (i=0; i<chans; i++) {
 		LMS_StopStream(&m_lms_stream_tx[i]);
 		LMS_StopStream(&m_lms_stream_rx[i]);
-
-		LMS_EnableChannel(m_lms_dev, LMS_CH_RX, i, false);
-		LMS_EnableChannel(m_lms_dev, LMS_CH_TX, i, false);
 	}
 
+	for (i=0; i<chans; i++) {
+		LMS_DestroyStream(m_lms_dev, &m_lms_stream_tx[i]);
+		LMS_DestroyStream(m_lms_dev, &m_lms_stream_rx[i]);
+	}
+
+	started = false;
 	return true;
 }
 
+/* do rx/tx calibration - depends on gain, freq and bw */
+bool LMSDevice::do_calib(size_t chan)
+{
+	LOGCHAN(chan, DDEV, INFO) << "Calibrating";
+	if (LMS_Calibrate(m_lms_dev, LMS_CH_RX, chan, LMS_CALIBRATE_BW_HZ, 0) < 0)
+		return false;
+	if (LMS_Calibrate(m_lms_dev, LMS_CH_TX, chan, LMS_CALIBRATE_BW_HZ, 0) < 0)
+		return false;
+	return true;
+}
+
+/* do rx/tx filter config - depends on bw only? */
+bool LMSDevice::do_filters(size_t chan)
+{
+	lms_range_t range_lpfbw_rx, range_lpfbw_tx;
+	float_type lpfbw_rx, lpfbw_tx;
+
+	LOGCHAN(chan, DDEV, INFO) << "Setting filters";
+	if (LMS_GetLPFBWRange(m_lms_dev, LMS_CH_RX, &range_lpfbw_rx))
+		return false;
+	print_range("LPFBWRange Rx", &range_lpfbw_rx);
+	if (LMS_GetLPFBWRange(m_lms_dev, LMS_CH_RX, &range_lpfbw_tx))
+		return false;
+	print_range("LPFBWRange Tx", &range_lpfbw_tx);
+
+	lpfbw_rx = OSMO_MIN(OSMO_MAX(1.4001e6, range_lpfbw_rx.min), range_lpfbw_rx.max);
+	lpfbw_tx = OSMO_MIN(OSMO_MAX(5.2e6, range_lpfbw_tx.min), range_lpfbw_tx.max);
+
+	LOGCHAN(chan, DDEV, INFO) << "LPFBW: Rx=" << lpfbw_rx << " Tx=" << lpfbw_tx;
+
+	LOGCHAN(chan, DDEV, INFO) << "Setting LPFBW";
+	if (LMS_SetLPFBW(m_lms_dev, LMS_CH_RX, chan, lpfbw_rx) < 0)
+		return false;
+	if (LMS_SetLPFBW(m_lms_dev, LMS_CH_TX, chan, lpfbw_tx) < 0)
+		return false;
+	return true;
+}
+
+
 double LMSDevice::maxTxGain()
 {
-	return 73.0;
+	return maxTxGainClamp;
 }
 
 double LMSDevice::minTxGain()
@@ -302,40 +405,30 @@ double LMSDevice::minRxGain()
 
 double LMSDevice::setTxGain(double dB, size_t chan)
 {
-	if (chan) {
-		LOG(ALERT) << "Invalid channel " << chan;
-		return 0.0;
-	}
-
 	if (dB > maxTxGain())
 		dB = maxTxGain();
 	if (dB < minTxGain())
 		dB = minTxGain();
 
-	LOG(NOTICE) << "Setting TX gain to " << dB << " dB.";
+	LOGCHAN(chan, DDEV, NOTICE) << "Setting TX gain to " << dB << " dB";
 
 	if (LMS_SetGaindB(m_lms_dev, LMS_CH_TX, chan, dB) < 0)
-		LOG(ERR) << "Error setting TX gain";
+		LOGCHAN(chan, DDEV, ERR) << "Error setting TX gain to " << dB << " dB";
 
 	return dB;
 }
 
 double LMSDevice::setRxGain(double dB, size_t chan)
 {
-	if (chan) {
-		LOG(ALERT) << "Invalid channel " << chan;
-		return 0.0;
-	}
-
 	if (dB > maxRxGain())
 		dB = maxRxGain();
 	if (dB < minRxGain())
 		dB = minRxGain();
 
-	LOG(NOTICE) << "Setting RX gain to " << dB << " dB.";
+	LOGCHAN(chan, DDEV, NOTICE) << "Setting RX gain to " << dB << " dB";
 
 	if (LMS_SetGaindB(m_lms_dev, LMS_CH_RX, chan, dB) < 0)
-		LOG(ERR) << "Error setting RX gain";
+		LOGCHAN(chan, DDEV, ERR) << "Error setting RX gain to " << dB << " dB";
 
 	return dB;
 }
@@ -359,7 +452,7 @@ bool LMSDevice::flush_recv(size_t num_pkts)
 {
 	#define CHUNK 625
 	int len = CHUNK * tx_sps;
-	short *buffer = new short[len * 2];
+	short *buffer = (short*) alloca(sizeof(short) * len * 2);
 	int rc;
 	lms_stream_meta_t rx_metadata = {};
 	rx_metadata.flushPartialPacket = false;
@@ -369,18 +462,16 @@ bool LMSDevice::flush_recv(size_t num_pkts)
 
 	while (!ts_initial || (num_pkts-- > 0)) {
 		rc = LMS_RecvStream(&m_lms_stream_rx[0], &buffer[0], len, &rx_metadata, 100);
-		LOG(DEBUG) << "Flush: Recv buffer of len " << rc << " at " << std::hex << rx_metadata.timestamp;
+		LOGC(DDEV, DEBUG) << "Flush: Recv buffer of len " << rc << " at " << std::hex << rx_metadata.timestamp;
 		if (rc != len) {
-			LOG(ALERT) << "LMS: Device receive timed out";
-			delete[] buffer;
+			LOGC(DDEV, ERROR) << "Flush: Device receive timed out";
 			return false;
 		}
 
 		ts_initial = rx_metadata.timestamp + len;
 	}
 
-	LOG(INFO) << "Initial timestamp " << ts_initial << std::endl;
-	delete[] buffer;
+	LOGC(DDEV, INFO) << "Initial timestamp " << ts_initial << std::endl;
 	return true;
 }
 
@@ -389,18 +480,18 @@ bool LMSDevice::setRxAntenna(const std::string & ant, size_t chan)
 	int idx;
 
 	if (chan >= rx_paths.size()) {
-		LOG(ALERT) << "Requested non-existent channel " << chan;
+		LOGC(DDEV, ERROR) << "Requested non-existent channel " << chan;
 		return false;
 	}
 
 	idx = get_ant_idx(ant, LMS_CH_RX, chan);
 	if (idx < 0) {
-		LOG(ALERT) << "Invalid Rx Antenna";
+		LOGCHAN(chan, DDEV, ERROR) << "Invalid Rx Antenna";
 		return false;
 	}
 
 	if (LMS_SetAntenna(m_lms_dev, LMS_CH_RX, chan, idx) < 0) {
-		LOG(ALERT) << "Unable to set Rx Antenna";
+		LOGCHAN(chan, DDEV, ERROR) << "Unable to set Rx Antenna";
 	}
 
 	return true;
@@ -412,18 +503,18 @@ std::string LMSDevice::getRxAntenna(size_t chan)
 	int idx;
 
 	if (chan >= rx_paths.size()) {
-		LOG(ALERT) << "Requested non-existent channel " << chan;
+		LOGC(DDEV, ERROR) << "Requested non-existent channel " << chan;
 		return "";
 	}
 
 	idx = LMS_GetAntenna(m_lms_dev, LMS_CH_RX, chan);
 	if (idx < 0) {
-		LOG(ALERT) << "Error getting Rx Antenna";
+		LOGCHAN(chan, DDEV, ERROR) << "Error getting Rx Antenna";
 		return "";
 	}
 
 	if (LMS_GetAntennaList(m_lms_dev, LMS_CH_RX, chan, name_list) < idx) {
-		LOG(ALERT) << "Error getting Rx Antenna List";
+		LOGCHAN(chan, DDEV, ERROR) << "Error getting Rx Antenna List";
 		return "";
 	}
 
@@ -435,18 +526,18 @@ bool LMSDevice::setTxAntenna(const std::string & ant, size_t chan)
 	int idx;
 
 	if (chan >= tx_paths.size()) {
-		LOG(ALERT) << "Requested non-existent channel " << chan;
+		LOGC(DDEV, ERROR) << "Requested non-existent channel " << chan;
 		return false;
 	}
 
 	idx = get_ant_idx(ant, LMS_CH_TX, chan);
 	if (idx < 0) {
-		LOG(ALERT) << "Invalid Rx Antenna";
+		LOGCHAN(chan, DDEV, ERROR) << "Invalid Rx Antenna";
 		return false;
 	}
 
 	if (LMS_SetAntenna(m_lms_dev, LMS_CH_TX, chan, idx) < 0) {
-		LOG(ALERT) << "Unable to set Rx Antenna";
+		LOGCHAN(chan, DDEV, ERROR) << "Unable to set Rx Antenna";
 	}
 
 	return true;
@@ -458,18 +549,18 @@ std::string LMSDevice::getTxAntenna(size_t chan)
 	int idx;
 
 	if (chan >= tx_paths.size()) {
-		LOG(ALERT) << "Requested non-existent channel " << chan;
+		LOGC(DDEV, ERROR) << "Requested non-existent channel " << chan;
 		return "";
 	}
 
 	idx = LMS_GetAntenna(m_lms_dev, LMS_CH_TX, chan);
 	if (idx < 0) {
-		LOG(ALERT) << "Error getting Tx Antenna";
+		LOGCHAN(chan, DDEV, ERROR) << "Error getting Tx Antenna";
 		return "";
 	}
 
 	if (LMS_GetAntennaList(m_lms_dev, LMS_CH_TX, chan, name_list) < idx) {
-		LOG(ALERT) << "Error getting Tx Antenna List";
+		LOGCHAN(chan, DDEV, ERROR) << "Error getting Tx Antenna List";
 		return "";
 	}
 
@@ -482,66 +573,122 @@ bool LMSDevice::requiresRadioAlign()
 }
 
 GSM::Time LMSDevice::minLatency() {
-	/* Empirical data from a handful of
-	relatively recent machines shows that the B100 will underrun when
-	the transmit threshold is reduced to a time of 6 and a half frames,
-	so we set a minimum 7 frame threshold. */
-	return GSM::Time(6,7);
+	/* UNUSED on limesdr (only used on usrp1/2) */
+	return GSM::Time(0,0);
+}
+
+void LMSDevice::update_stream_stats(size_t chan, bool * underrun, bool * overrun)
+{
+	lms_stream_status_t status;
+	if (LMS_GetStreamStatus(&m_lms_stream_rx[chan], &status) == 0) {
+		if (status.underrun > m_last_rx_underruns[chan]) {
+			*underrun = true;
+			LOGCHAN(chan, DDEV, ERROR) << "recv Underrun! ("
+						   << m_last_rx_underruns[chan] << " -> "
+						   << status.underrun << ")";
+		}
+		m_last_rx_underruns[chan] = status.underrun;
+
+		if (status.overrun > m_last_rx_overruns[chan]) {
+			*overrun = true;
+			LOGCHAN(chan, DDEV, ERROR) << "recv Overrun! ("
+						   << m_last_rx_overruns[chan] << " -> "
+						   << status.overrun << ")";
+		}
+		m_last_rx_overruns[chan] = status.overrun;
+
+		if (status.droppedPackets > m_last_rx_dropped[chan]) {
+			LOGCHAN(chan, DDEV, ERROR) << "recv Dropped packets by HW! ("
+						   << m_last_rx_dropped[chan] << " -> "
+						   << status.droppedPackets << ")";
+		}
+		m_last_rx_dropped[chan] = m_last_rx_overruns[chan];
+	}
 }
 
 // NOTE: Assumes sequential reads
 int LMSDevice::readSamples(std::vector < short *>&bufs, int len, bool * overrun,
 			   TIMESTAMP timestamp, bool * underrun, unsigned *RSSI)
 {
-	int rc = 0;
+	int rc, num_smpls, expect_smpls;
+	ssize_t avail_smpls;
+	TIMESTAMP expect_timestamp;
 	unsigned int i;
-	lms_stream_status_t status;
 	lms_stream_meta_t rx_metadata = {};
 	rx_metadata.flushPartialPacket = false;
 	rx_metadata.waitForTimestamp = false;
 	rx_metadata.timestamp = 0;
 
 	if (bufs.size() != chans) {
-		LOG(ALERT) << "Invalid channel combination " << bufs.size();
+		LOGC(DDEV, ERROR) << "Invalid channel combination " << bufs.size();
 		return -1;
 	}
 
 	*overrun = false;
 	*underrun = false;
-	for (i = 0; i<chans; i++) {
-		thread_enable_cancel(false);
-		rc = LMS_RecvStream(&m_lms_stream_rx[i], bufs[i], len, &rx_metadata, 100);
-		if (timestamp != (TIMESTAMP)rx_metadata.timestamp)
-			LOG(ALERT) << "chan "<< i << " recv buffer of len " << rc << " expect " << std::hex << timestamp << " got " << std::hex << (TIMESTAMP)rx_metadata.timestamp << " (" << std::hex << rx_metadata.timestamp <<") diff=" << rx_metadata.timestamp - timestamp;
-		if (rc != len) {
-			LOG(ALERT) << "LMS: Device receive timed out";
-		}
 
-		if (LMS_GetStreamStatus(&m_lms_stream_rx[i], &status) == 0) {
-			if (status.underrun > m_last_rx_underruns[i])
-				*underrun = true;
-			m_last_rx_underruns[i] = status.underrun;
-
-			if (status.overrun > m_last_rx_overruns[i])
-				*overrun = true;
-			m_last_rx_overruns[i] = status.overrun;
-		}
-		thread_enable_cancel(true);
+	/* Check that timestamp is valid */
+	rc = rx_buffers[0]->avail_smpls(timestamp);
+	if (rc < 0) {
+		LOGC(DDEV, ERROR) << rx_buffers[0]->str_code(rc);
+		LOGC(DDEV, ERROR) << rx_buffers[0]->str_status(timestamp);
+		return 0;
 	}
 
-	samplesRead += rc;
+	for (i = 0; i<chans; i++) {
+		/* Receive samples from HW until we have enough */
+		while ((avail_smpls = rx_buffers[i]->avail_smpls(timestamp)) < len) {
+			thread_enable_cancel(false);
+			num_smpls = LMS_RecvStream(&m_lms_stream_rx[i], bufs[i], len - avail_smpls, &rx_metadata, 100);
+			update_stream_stats(i, underrun, overrun);
+			thread_enable_cancel(true);
+			if (num_smpls <= 0) {
+				LOGCHAN(i, DDEV, ERROR) << "Device receive timed out (" << rc << " vs exp " << len << ").";
+				return -1;
+			}
 
-	if (((TIMESTAMP) rx_metadata.timestamp) < timestamp)
-		rc = 0;
+			LOGCHAN(i, DDEV, DEBUG) "Received timestamp = " << (TIMESTAMP)rx_metadata.timestamp << " (" << num_smpls << ")";
 
-	return rc;
+			expect_smpls = len - avail_smpls;
+			if (expect_smpls != num_smpls)
+				LOGCHAN(i, DDEV, NOTICE) << "Unexpected recv buffer len: expect "
+							 << expect_smpls << " got " << num_smpls
+							 << ", diff=" << expect_smpls - num_smpls;
+
+			expect_timestamp = timestamp + avail_smpls;
+			if (expect_timestamp != (TIMESTAMP)rx_metadata.timestamp)
+				LOGCHAN(i, DDEV, ERROR) << "Unexpected recv buffer timestamp: expect "
+							<< expect_timestamp << " got " << (TIMESTAMP)rx_metadata.timestamp
+							<< ", diff=" << rx_metadata.timestamp - expect_timestamp;
+
+			rc = rx_buffers[i]->write(bufs[i], num_smpls, (TIMESTAMP)rx_metadata.timestamp);
+			if (rc < 0) {
+				LOGCHAN(i, DDEV, ERROR) << rx_buffers[i]->str_code(rc);
+				LOGCHAN(i, DDEV, ERROR) << rx_buffers[i]->str_status(timestamp);
+				if (rc != smpl_buf::ERROR_OVERFLOW)
+					return 0;
+			}
+		}
+	}
+
+	/* We have enough samples */
+	for (size_t i = 0; i < rx_buffers.size(); i++) {
+		rc = rx_buffers[i]->read(bufs[i], len, timestamp);
+		if ((rc < 0) || (rc != len)) {
+			LOGC(DDEV, ERROR) << rx_buffers[i]->str_code(rc);
+			LOGC(DDEV, ERROR) << rx_buffers[i]->str_status(timestamp);
+			return 0;
+		}
+	}
+
+	return len;
 }
 
 int LMSDevice::writeSamples(std::vector < short *>&bufs, int len,
 			    bool * underrun, unsigned long long timestamp,
 			    bool isControl)
 {
-	int rc;
+	int rc = 0;
 	unsigned int i;
 	lms_stream_status_t status;
 	lms_stream_meta_t tx_metadata = {};
@@ -550,23 +697,23 @@ int LMSDevice::writeSamples(std::vector < short *>&bufs, int len,
 	tx_metadata.timestamp = timestamp - ts_offset;	/* Shift Tx time by offset */
 
 	if (isControl) {
-		LOG(ERR) << "Control packets not supported";
+		LOGC(DDEV, ERROR) << "Control packets not supported";
 		return 0;
 	}
 
 	if (bufs.size() != chans) {
-		LOG(ALERT) << "Invalid channel combination " << bufs.size();
+		LOGC(DDEV, ERROR) << "Invalid channel combination " << bufs.size();
 		return -1;
 	}
 
 	*underrun = false;
 
 	for (i = 0; i<chans; i++) {
-		LOG(DEBUG) << "chan "<< i << " send buffer of len " << len << " timestamp " << std::hex << tx_metadata.timestamp;
+		LOGCHAN(i, DDEV, DEBUG) << "send buffer of len " << len << " timestamp " << std::hex << tx_metadata.timestamp;
 		thread_enable_cancel(false);
 		rc = LMS_SendStream(&m_lms_stream_tx[i], bufs[i], len, &tx_metadata, 100);
 		if (rc != len) {
-			LOG(ALERT) << "LMS: Device send timed out";
+			LOGCHAN(i, DDEV, ERROR) << "LMS: Device send timed out";
 		}
 
 		if (LMS_GetStreamStatus(&m_lms_stream_tx[i], &status) == 0) {
@@ -577,8 +724,6 @@ int LMSDevice::writeSamples(std::vector < short *>&bufs, int len,
 		thread_enable_cancel(true);
 	}
 
-	samplesWritten += rc;
-
 	return rc;
 }
 
@@ -589,14 +734,10 @@ bool LMSDevice::updateAlignment(TIMESTAMP timestamp)
 
 bool LMSDevice::setTxFreq(double wFreq, size_t chan)
 {
-
-	if (chan) {
-		LOG(ALERT) << "Invalid channel " << chan;
-		return false;
-	}
+	LOGCHAN(chan, DDEV, NOTICE) << "Setting Tx Freq to " << wFreq << " Hz";
 
 	if (LMS_SetLOFrequency(m_lms_dev, LMS_CH_TX, chan, wFreq) < 0) {
-		LOG(ALERT) << "set Tx: " << wFreq << " failed!";
+		LOGCHAN(chan, DDEV, ERROR) << "Error setting Tx Freq to " << wFreq << " Hz";
 		return false;
 	}
 
@@ -605,13 +746,10 @@ bool LMSDevice::setTxFreq(double wFreq, size_t chan)
 
 bool LMSDevice::setRxFreq(double wFreq, size_t chan)
 {
-	if (chan) {
-		LOG(ALERT) << "Invalid channel " << chan;
-		return false;
-	}
+	LOGCHAN(chan, DDEV, NOTICE) << "Setting Rx Freq to " << wFreq << " Hz";
 
 	if (LMS_SetLOFrequency(m_lms_dev, LMS_CH_RX, chan, wFreq) < 0) {
-		LOG(ALERT) << "set Rx: " << wFreq << " failed!";
+		LOGCHAN(chan, DDEV, ERROR) << "Error setting Rx Freq to " << wFreq << " Hz";
 		return false;
 	}
 
@@ -624,11 +762,11 @@ RadioDevice *RadioDevice::make(size_t tx_sps, size_t rx_sps,
 			       const std::vector < std::string > &rx_paths)
 {
 	if (tx_sps != rx_sps) {
-		LOG(ERROR) << "LMS Requires tx_sps == rx_sps";
+		LOGC(DDEV, ERROR) << "LMS Requires tx_sps == rx_sps";
 		return NULL;
 	}
 	if (lo_offset != 0.0) {
-		LOG(ERROR) << "LMS doesn't support lo_offset";
+		LOGC(DDEV, ERROR) << "LMS doesn't support lo_offset";
 		return NULL;
 	}
 	return new LMSDevice(tx_sps, rx_sps, iface, chans, lo_offset, tx_paths, rx_paths);
