@@ -23,12 +23,9 @@
 
 #include <map>
 #include "radioDevice.h"
+#include "UHDDevice.h"
 #include "Threads.h"
 #include "Logger.h"
-#include <uhd/version.hpp>
-#include <uhd/property_tree.hpp>
-#include <uhd/usrp/multi_usrp.hpp>
-#include <uhd/utils/thread_priority.hpp>
 
 #ifdef HAVE_CONFIG_H
 #include "config.h"
@@ -36,6 +33,9 @@
 
 #ifndef USE_UHD_3_11
 #include <uhd/utils/msg.hpp>
+#include <uhd/utils/thread_priority.hpp>
+#else
+#include <uhd/utils/thread.hpp>
 #endif
 
 #define USRP_TX_AMPL     0.3
@@ -56,20 +56,6 @@
  */
 #define UMTRX_VGA1_DEF   -18
 
-enum uhd_dev_type {
-	USRP1,
-	USRP2,
-	B100,
-	B200,
-	B210,
-	B2XX_MCBTS,
-	E1XX,
-	E3XX,
-	X3XX,
-	UMTRX,
-	LIMESDR,
-};
-
 /*
  * USRP version dependent device timings
  */
@@ -134,192 +120,9 @@ static const std::map<dev_key, dev_desc> dev_param_map {
 	{ std::make_tuple(B2XX_MCBTS, 4, 4), { 1, 51.2e6, MCBTS_SPACING*4, B2XX_TIMING_MCBTS, "B200/B210 4 SPS Multi-ARFCN" } },
 };
 
-/*
-    Sample Buffer - Allows reading and writing of timed samples using osmo-trx
-                    or UHD style timestamps. Time conversions are handled
-                    internally or accessable through the static convert calls.
-*/
-class smpl_buf {
-public:
-	/** Sample buffer constructor
-	    @param len number of 32-bit samples the buffer should hold
-	    @param rate sample clockrate
-	    @param timestamp
-	*/
-	smpl_buf(size_t len, double rate);
-	~smpl_buf();
-
-	/** Query number of samples available for reading
-	    @param timestamp time of first sample
-	    @return number of available samples or error
-	*/
-	ssize_t avail_smpls(TIMESTAMP timestamp) const;
-	ssize_t avail_smpls(uhd::time_spec_t timestamp) const;
-
-	/** Read and write
-	    @param buf pointer to buffer
-	    @param len number of samples desired to read or write
-	    @param timestamp time of first stample
-	    @return number of actual samples read or written or error
-	*/
-	ssize_t read(void *buf, size_t len, TIMESTAMP timestamp);
-	ssize_t read(void *buf, size_t len, uhd::time_spec_t timestamp);
-	ssize_t write(void *buf, size_t len, TIMESTAMP timestamp);
-	ssize_t write(void *buf, size_t len, uhd::time_spec_t timestamp);
-
-	/** Buffer status string
-	    @return a formatted string describing internal buffer state
-	*/
-	std::string str_status(size_t ts) const;
-
-	/** Formatted error string
-	    @param code an error code
-	    @return a formatted error string
-	*/
-	static std::string str_code(ssize_t code);
-
-	enum err_code {
-		ERROR_TIMESTAMP = -1,
-		ERROR_READ = -2,
-		ERROR_WRITE = -3,
-		ERROR_OVERFLOW = -4
-	};
-
-private:
-	uint32_t *data;
-	size_t buf_len;
-
-	double clk_rt;
-
-	TIMESTAMP time_start;
-	TIMESTAMP time_end;
-
-	size_t data_start;
-	size_t data_end;
-};
-
-/*
-    uhd_device - UHD implementation of the Device interface. Timestamped samples
-                are sent to and received from the device. An intermediate buffer
-                on the receive side collects and aligns packets of samples.
-                Events and errors such as underruns are reported asynchronously
-                by the device and received in a separate thread.
-*/
-class uhd_device : public RadioDevice {
-public:
-	uhd_device(size_t tx_sps, size_t rx_sps, InterfaceType type,
-		   size_t chans, double offset,
-		   const std::vector<std::string>& tx_paths,
-		   const std::vector<std::string>& rx_paths);
-	~uhd_device();
-
-	int open(const std::string &args, int ref, bool swap_channels);
-	bool start();
-	bool stop();
-	bool restart();
-	void setPriority(float prio);
-	enum TxWindowType getWindowType() { return tx_window; }
-
-	int readSamples(std::vector<short *> &bufs, int len, bool *overrun,
-			TIMESTAMP timestamp, bool *underrun, unsigned *RSSI);
-
-	int writeSamples(std::vector<short *> &bufs, int len, bool *underrun,
-			 TIMESTAMP timestamp, bool isControl);
-
-	bool updateAlignment(TIMESTAMP timestamp);
-
-	bool setTxFreq(double wFreq, size_t chan);
-	bool setRxFreq(double wFreq, size_t chan);
-
-	TIMESTAMP initialWriteTimestamp();
-	TIMESTAMP initialReadTimestamp();
-
-	double fullScaleInputValue();
-	double fullScaleOutputValue();
-
-	double setRxGain(double db, size_t chan);
-	double getRxGain(size_t chan);
-	double maxRxGain(void) { return rx_gain_max; }
-	double minRxGain(void) { return rx_gain_min; }
-
-	double setTxGain(double db, size_t chan);
-	double maxTxGain(void) { return tx_gain_max; }
-	double minTxGain(void) { return tx_gain_min; }
-
-	double getTxFreq(size_t chan);
-	double getRxFreq(size_t chan);
-	double getRxFreq();
-
-	bool setRxAntenna(const std::string &ant, size_t chan);
-	std::string getRxAntenna(size_t chan);
-	bool setTxAntenna(const std::string &ant, size_t chan);
-	std::string getTxAntenna(size_t chan);
-
-	bool requiresRadioAlign();
-
-	GSM::Time minLatency();
-
-	inline double getSampleRate() { return tx_rate; }
-	inline double numberRead() { return rx_pkt_cnt; }
-	inline double numberWritten() { return 0; }
-
-	/** Receive and process asynchronous message
-	    @return true if message received or false on timeout or error
-	*/
-	bool recv_async_msg();
-
-	enum err_code {
-		ERROR_TIMING = -1,
-		ERROR_TIMEOUT = -2,
-		ERROR_UNRECOVERABLE = -3,
-		ERROR_UNHANDLED = -4,
-	};
-
-private:
-	uhd::usrp::multi_usrp::sptr usrp_dev;
-	uhd::tx_streamer::sptr tx_stream;
-	uhd::rx_streamer::sptr rx_stream;
-	enum TxWindowType tx_window;
-	enum uhd_dev_type dev_type;
-
-	double tx_rate, rx_rate;
-
-	double tx_gain_min, tx_gain_max;
-	double rx_gain_min, rx_gain_max;
-
-	std::vector<double> tx_gains, rx_gains;
-	std::vector<double> tx_freqs, rx_freqs;
-	size_t tx_spp, rx_spp;
-
-	bool started;
-	bool aligned;
-
-	size_t rx_pkt_cnt;
-	size_t drop_cnt;
-	uhd::time_spec_t prev_ts;
-
-	TIMESTAMP ts_initial, ts_offset;
-	std::vector<smpl_buf *> rx_buffers;
-
-	void init_gains();
-	void set_channels(bool swap);
-	void set_rates();
-	bool parse_dev_type();
-	bool flush_recv(size_t num_pkts);
-	int check_rx_md_err(uhd::rx_metadata_t &md, ssize_t num_smpls);
-
-	std::string str_code(uhd::rx_metadata_t metadata);
-	std::string str_code(uhd::async_metadata_t metadata);
-
-	uhd::tune_request_t select_freq(double wFreq, size_t chan, bool tx);
-	bool set_freq(double freq, size_t chan, bool tx);
-
-	Thread *async_event_thrd;
-	Mutex tune_lock;
-};
-
 void *async_event_loop(uhd_device *dev)
 {
+	set_selfthread_name("UHDAsyncEvent");
 	dev->setPriority(0.43);
 
 	while (1) {
@@ -340,13 +143,13 @@ void uhd_msg_handler(uhd::msg::type_t type, const std::string &msg)
 {
 	switch (type) {
 	case uhd::msg::status:
-		LOG(INFO) << msg;
+		LOGC(DDEV, INFO) << msg;
 		break;
 	case uhd::msg::warning:
-		LOG(WARNING) << msg;
+		LOGC(DDEV, WARNING) << msg;
 		break;
 	case uhd::msg::error:
-		LOG(ERR) << msg;
+		LOGC(DDEV, ERROR) << msg;
 		break;
 	case uhd::msg::fastpath:
 		break;
@@ -354,12 +157,6 @@ void uhd_msg_handler(uhd::msg::type_t type, const std::string &msg)
 }
 #endif
 
-static void thread_enable_cancel(bool cancel)
-{
-	cancel ? pthread_setcancelstate(PTHREAD_CANCEL_ENABLE, NULL) :
-		 pthread_setcancelstate(PTHREAD_CANCEL_DISABLE, NULL);
-}
-
 uhd_device::uhd_device(size_t tx_sps, size_t rx_sps,
 		       InterfaceType iface, size_t chans, double lo_offset,
 		       const std::vector<std::string>& tx_paths,
@@ -368,7 +165,7 @@ uhd_device::uhd_device(size_t tx_sps, size_t rx_sps,
 	  tx_gain_min(0.0), tx_gain_max(0.0),
 	  rx_gain_min(0.0), rx_gain_max(0.0),
 	  tx_spp(0), rx_spp(0),
-	  started(false), aligned(false), rx_pkt_cnt(0), drop_cnt(0),
+	  started(false), aligned(false), drop_cnt(0),
 	  prev_ts(0,0), ts_initial(0), ts_offset(0), async_event_thrd(NULL)
 {
 }
@@ -388,7 +185,7 @@ void uhd_device::init_gains()
 	if (dev_type == UMTRX) {
 		std::vector<std::string> gain_stages = usrp_dev->get_tx_gain_names(0);
 		if (gain_stages[0] == "VGA") {
-			LOG(WARNING) << "Update your UHD version for a proper Tx gain support";
+			LOGC(DDEV, WARNING) << "Update your UHD version for a proper Tx gain support";
 		}
 		if (gain_stages[0] == "VGA" || gain_stages[0] == "PA") {
 			range = usrp_dev->get_tx_gain_range();
@@ -404,23 +201,23 @@ void uhd_device::init_gains()
 		tx_gain_min = range.start();
 		tx_gain_max = range.stop();
 	}
-	LOG(INFO) << "Supported Tx gain range [" << tx_gain_min << "; " << tx_gain_max << "]";
+	LOGC(DDEV, INFO) << "Supported Tx gain range [" << tx_gain_min << "; " << tx_gain_max << "]";
 
 	range = usrp_dev->get_rx_gain_range();
 	rx_gain_min = range.start();
 	rx_gain_max = range.stop();
-	LOG(INFO) << "Supported Rx gain range [" << rx_gain_min << "; " << rx_gain_max << "]";
+	LOGC(DDEV, INFO) << "Supported Rx gain range [" << rx_gain_min << "; " << rx_gain_max << "]";
 
 	for (size_t i = 0; i < tx_gains.size(); i++) {
 		double gain = (tx_gain_min + tx_gain_max) / 2;
-		LOG(INFO) << "Default setting Tx gain for channel " << i << " to " << gain;
+		LOGC(DDEV, INFO) << "Default setting Tx gain for channel " << i << " to " << gain;
 		usrp_dev->set_tx_gain(gain, i);
 		tx_gains[i] = usrp_dev->get_tx_gain(i);
 	}
 
 	for (size_t i = 0; i < rx_gains.size(); i++) {
 		double gain = (rx_gain_min + rx_gain_max) / 2;
-		LOG(INFO) << "Default setting Rx gain for channel " << i << " to " << gain;
+		LOGC(DDEV, INFO) << "Default setting Rx gain for channel " << i << " to " << gain;
 		usrp_dev->set_rx_gain(gain, i);
 		rx_gains[i] = usrp_dev->get_rx_gain(i);
 	}
@@ -444,7 +241,7 @@ void uhd_device::set_rates()
 	rx_rate = usrp_dev->get_rx_rate();
 
 	ts_offset = static_cast<TIMESTAMP>(desc.offset * rx_rate);
-	LOG(INFO) << "Rates configured for " << desc.str;
+	LOGC(DDEV, INFO) << "Rates configured for " << desc.str;
 }
 
 double uhd_device::setTxGain(double db, size_t chan)
@@ -453,7 +250,7 @@ double uhd_device::setTxGain(double db, size_t chan)
 		chan = 0;
 
 	if (chan >= tx_gains.size()) {
-		LOG(ALERT) << "Requested non-existent channel" << chan;
+		LOGC(DDEV, ALERT) << "Requested non-existent channel" << chan;
 		return 0.0f;
 	}
 
@@ -476,22 +273,25 @@ double uhd_device::setTxGain(double db, size_t chan)
 
 	tx_gains[chan] = usrp_dev->get_tx_gain(chan);
 
-	LOG(INFO) << "Set TX gain to " << tx_gains[chan] << "dB (asked for " << db << "dB)";
+	LOGC(DDEV, INFO) << "Set TX gain to " << tx_gains[chan] << "dB (asked for " << db << "dB)";
 
 	return tx_gains[chan];
 }
 
 double uhd_device::setRxGain(double db, size_t chan)
 {
+	if (iface == MULTI_ARFCN)
+		chan = 0;
+
 	if (chan >= rx_gains.size()) {
-		LOG(ALERT) << "Requested non-existent channel " << chan;
+		LOGC(DDEV, ALERT) << "Requested non-existent channel " << chan;
 		return 0.0f;
 	}
 
 	usrp_dev->set_rx_gain(db, chan);
 	rx_gains[chan] = usrp_dev->get_rx_gain(chan);
 
-	LOG(INFO) << "Set RX gain to " << rx_gains[chan] << "dB (asked for " << db << "dB)";
+	LOGC(DDEV, INFO) << "Set RX gain to " << rx_gains[chan] << "dB (asked for " << db << "dB)";
 
 	return rx_gains[chan];
 }
@@ -502,7 +302,7 @@ double uhd_device::getRxGain(size_t chan)
 		chan = 0;
 
 	if (chan >= rx_gains.size()) {
-		LOG(ALERT) << "Requested non-existent channel " << chan;
+		LOGC(DDEV, ALERT) << "Requested non-existent channel " << chan;
 		return 0.0f;
 	}
 
@@ -550,7 +350,7 @@ bool uhd_device::parse_dev_type()
 		mapIter++;
 	}
 
-	LOG(ALERT) << "Unsupported device " << devString;
+	LOGC(DDEV, ALERT) << "Unsupported device " << devString;
 	return false;
 }
 
@@ -621,16 +421,16 @@ int uhd_device::open(const std::string &args, int ref, bool swap_channels)
 	uhd::device_addr_t addr(args);
 	uhd::device_addrs_t dev_addrs = uhd::device::find(addr);
 	if (dev_addrs.size() == 0) {
-		LOG(ALERT) << "No UHD devices found with address '" << args << "'";
+		LOGC(DDEV, ALERT) << "No UHD devices found with address '" << args << "'";
 		return -1;
 	}
 
 	// Use the first found device
-	LOG(INFO) << "Using discovered UHD device " << dev_addrs[0].to_string();
+	LOGC(DDEV, INFO) << "Using discovered UHD device " << dev_addrs[0].to_string();
 	try {
 		usrp_dev = uhd::usrp::multi_usrp::make(addr);
-	} catch(...) {
-		LOG(ALERT) << "UHD make failed, device " << args;
+	} catch(uhd::key_error::exception &e) {
+		LOGC(DDEV, ALERT) << "UHD make failed, device " << args << ", exception:\n" << e.what();
 		return -1;
 	}
 
@@ -639,19 +439,19 @@ int uhd_device::open(const std::string &args, int ref, bool swap_channels)
 		return -1;
 
 	if ((dev_type == E3XX) && !uhd_e3xx_version_chk()) {
-		LOG(ALERT) << "E3XX requires UHD 003.009.000 or greater";
+		LOGC(DDEV, ALERT) << "E3XX requires UHD 003.009.000 or greater";
 		return -1;
 	}
 
 	try {
 		set_channels(swap_channels);
         } catch (const std::exception &e) {
-		LOG(ALERT) << "Channel setting failed - " << e.what();
+		LOGC(DDEV, ALERT) << "Channel setting failed - " << e.what();
 		return -1;
 	}
 
 	if (!set_antennas()) {
-		LOG(ALERT) << "UHD antenna setting failed";
+		LOGC(DDEV, ALERT) << "UHD antenna setting failed";
 		return -1;
 	}
 
@@ -672,7 +472,7 @@ int uhd_device::open(const std::string &args, int ref, bool swap_channels)
 		refstr = "gpsdo";
 		break;
 	default:
-		LOG(ALERT) << "Invalid reference type";
+		LOGC(DDEV, ALERT) << "Invalid reference type";
 		return -1;
 	}
 
@@ -681,7 +481,7 @@ int uhd_device::open(const std::string &args, int ref, bool swap_channels)
 	try {
 		set_rates();
         } catch (const std::exception &e) {
-		LOG(ALERT) << "UHD rate setting failed - " << e.what();
+		LOGC(DDEV, ALERT) << "UHD rate setting failed - " << e.what();
 		return -1;
 	}
 
@@ -714,13 +514,18 @@ int uhd_device::open(const std::string &args, int ref, bool swap_channels)
 	// Create receive buffer
 	size_t buf_len = SAMPLE_BUF_SZ / sizeof(uint32_t);
 	for (size_t i = 0; i < rx_buffers.size(); i++)
-		rx_buffers[i] = new smpl_buf(buf_len, rx_rate);
+		rx_buffers[i] = new smpl_buf(buf_len);
+
+	// Create vector buffer
+	pkt_bufs = std::vector<std::vector<short> >(chans, std::vector<short>(2 * rx_spp));
+	for (size_t i = 0; i < pkt_bufs.size(); i++)
+		pkt_ptrs.push_back(&pkt_bufs[i].front());
 
 	// Initialize and shadow gain values
 	init_gains();
 
 	// Print configuration
-	LOG(INFO) << "\n" << usrp_dev->get_pp_string();
+	LOGC(DDEV, INFO) << "\n" << usrp_dev->get_pp_string();
 
 	if (iface == MULTI_ARFCN)
 		return MULTI_ARFCN;
@@ -749,13 +554,6 @@ bool uhd_device::flush_recv(size_t num_pkts)
 	size_t num_smpls;
 	float timeout = UHD_RESTART_TIMEOUT;
 
-	std::vector<std::vector<short> >
-		pkt_bufs(chans, std::vector<short>(2 * rx_spp));
-
-	std::vector<short *> pkt_ptrs;
-	for (size_t i = 0; i < pkt_bufs.size(); i++)
-		pkt_ptrs.push_back(&pkt_bufs[i].front());
-
 	ts_initial = 0;
 	while (!ts_initial || (num_pkts-- > 0)) {
 		num_smpls = rx_stream->recv(pkt_ptrs, rx_spp, md,
@@ -763,7 +561,7 @@ bool uhd_device::flush_recv(size_t num_pkts)
 		if (!num_smpls) {
 			switch (md.error_code) {
 			case uhd::rx_metadata_t::ERROR_CODE_TIMEOUT:
-				LOG(ALERT) << "Device timed out";
+				LOGC(DDEV, ALERT) << "Device timed out";
 				return false;
 			default:
 				continue;
@@ -773,7 +571,7 @@ bool uhd_device::flush_recv(size_t num_pkts)
 		ts_initial = md.time_spec.to_ticks(rx_rate);
 	}
 
-	LOG(INFO) << "Initial timestamp " << ts_initial << std::endl;
+	LOGC(DDEV, INFO) << "Initial timestamp " << ts_initial << std::endl;
 
 	return true;
 }
@@ -798,10 +596,10 @@ bool uhd_device::restart()
 
 bool uhd_device::start()
 {
-	LOG(INFO) << "Starting USRP...";
+	LOGC(DDEV, INFO) << "Starting USRP...";
 
 	if (started) {
-		LOG(ERR) << "Device already started";
+		LOGC(DDEV, ERROR) << "Device already started";
 		return false;
 	}
 
@@ -819,7 +617,7 @@ bool uhd_device::start()
 
 	// Display usrp time
 	double time_now = usrp_dev->get_time_now().get_real_secs();
-	LOG(INFO) << "The current time is " << time_now << " seconds";
+	LOGC(DDEV, INFO) << "The current time is " << time_now << " seconds";
 
 	started = true;
 	return true;
@@ -852,11 +650,11 @@ void uhd_device::setPriority(float prio)
 int uhd_device::check_rx_md_err(uhd::rx_metadata_t &md, ssize_t num_smpls)
 {
 	if (!num_smpls) {
-		LOG(ERR) << str_code(md);
+		LOGC(DDEV, ERROR) << str_code(md);
 
 		switch (md.error_code) {
 		case uhd::rx_metadata_t::ERROR_CODE_TIMEOUT:
-			LOG(ALERT) << "UHD: Receive timed out";
+			LOGC(DDEV, ALERT) << "UHD: Receive timed out";
 			return ERROR_TIMEOUT;
 		case uhd::rx_metadata_t::ERROR_CODE_OVERFLOW:
 		case uhd::rx_metadata_t::ERROR_CODE_LATE_COMMAND:
@@ -869,14 +667,14 @@ int uhd_device::check_rx_md_err(uhd::rx_metadata_t &md, ssize_t num_smpls)
 
 	// Missing timestamp
 	if (!md.has_time_spec) {
-		LOG(ALERT) << "UHD: Received packet missing timestamp";
+		LOGC(DDEV, ALERT) << "UHD: Received packet missing timestamp";
 		return ERROR_UNRECOVERABLE;
 	}
 
 	// Monotonicity check
 	if (md.time_spec < prev_ts) {
-		LOG(ALERT) << "UHD: Loss of monotonic time";
-		LOG(ALERT) << "Current time: " << md.time_spec.get_real_secs() << ", "
+		LOGC(DDEV, ALERT) << "UHD: Loss of monotonic time";
+		LOGC(DDEV, ALERT) << "Current time: " << md.time_spec.get_real_secs() << ", "
 			   << "Previous time: " << prev_ts.get_real_secs();
 		return ERROR_TIMING;
 	}
@@ -899,7 +697,7 @@ int uhd_device::readSamples(std::vector<short *> &bufs, int len, bool *overrun,
 	uhd::rx_metadata_t metadata;
 
 	if (bufs.size() != chans) {
-		LOG(ALERT) << "Invalid channel combination " << bufs.size();
+		LOGC(DDEV, ALERT) << "Invalid channel combination " << bufs.size();
 		return -1;
 	}
 
@@ -910,24 +708,16 @@ int uhd_device::readSamples(std::vector<short *> &bufs, int len, bool *overrun,
 	timestamp += ts_offset;
 
 	ts = uhd::time_spec_t::from_ticks(timestamp, rx_rate);
-	LOG(DEBUG) << "Requested timestamp = " << ts.get_real_secs();
+	LOGC(DDEV, DEBUG) << "Requested timestamp = " << ts.get_real_secs();
 
 	// Check that timestamp is valid
 	rc = rx_buffers[0]->avail_smpls(timestamp);
 	if (rc < 0) {
-		LOG(ERR) << rx_buffers[0]->str_code(rc);
-		LOG(ERR) << rx_buffers[0]->str_status(timestamp);
+		LOGC(DDEV, ERROR) << rx_buffers[0]->str_code(rc);
+		LOGC(DDEV, ERROR) << rx_buffers[0]->str_status(timestamp);
 		return 0;
 	}
 
-	// Create vector buffer
-	std::vector<std::vector<short> >
-		pkt_bufs(chans, std::vector<short>(2 * rx_spp));
-
-	std::vector<short *> pkt_ptrs;
-	for (size_t i = 0; i < pkt_bufs.size(); i++)
-		pkt_ptrs.push_back(&pkt_bufs[i].front());
-
 	// Receive samples from the usrp until we have enough
 	while (rx_buffers[0]->avail_smpls(timestamp) < len) {
 		thread_enable_cancel(false);
@@ -935,14 +725,12 @@ int uhd_device::readSamples(std::vector<short *> &bufs, int len, bool *overrun,
 						   metadata, 0.1, true);
 		thread_enable_cancel(true);
 
-		rx_pkt_cnt++;
-
 		// Check for errors
 		rc = check_rx_md_err(metadata, num_smpls);
 		switch (rc) {
 		case ERROR_UNRECOVERABLE:
-			LOG(ALERT) << "UHD: Version " << uhd::get_version_string();
-			LOG(ALERT) << "UHD: Unrecoverable error, exiting...";
+			LOGC(DDEV, ALERT) << "UHD: Version " << uhd::get_version_string();
+			LOGC(DDEV, ALERT) << "UHD: Unrecoverable error, exiting...";
 			exit(-1);
 		case ERROR_TIMEOUT:
 			// Assume stopping condition
@@ -954,17 +742,17 @@ int uhd_device::readSamples(std::vector<short *> &bufs, int len, bool *overrun,
 		}
 
 		ts = metadata.time_spec;
-		LOG(DEBUG) << "Received timestamp = " << ts.get_real_secs();
+		LOGC(DDEV, DEBUG) << "Received timestamp = " << ts.get_real_secs();
 
 		for (size_t i = 0; i < rx_buffers.size(); i++) {
 			rc = rx_buffers[i]->write((short *) &pkt_bufs[i].front(),
 						  num_smpls,
-						  metadata.time_spec);
+						  metadata.time_spec.to_ticks(rx_rate));
 
 			// Continue on local overrun, exit on other errors
 			if ((rc < 0)) {
-				LOG(ERR) << rx_buffers[i]->str_code(rc);
-				LOG(ERR) << rx_buffers[i]->str_status(timestamp);
+				LOGC(DDEV, ERROR) << rx_buffers[i]->str_code(rc);
+				LOGC(DDEV, ERROR) << rx_buffers[i]->str_status(timestamp);
 				if (rc != smpl_buf::ERROR_OVERFLOW)
 					return 0;
 			}
@@ -975,8 +763,8 @@ int uhd_device::readSamples(std::vector<short *> &bufs, int len, bool *overrun,
 	for (size_t i = 0; i < rx_buffers.size(); i++) {
 		rc = rx_buffers[i]->read(bufs[i], len, timestamp);
 		if ((rc < 0) || (rc != len)) {
-			LOG(ERR) << rx_buffers[i]->str_code(rc);
-			LOG(ERR) << rx_buffers[i]->str_status(timestamp);
+			LOGC(DDEV, ERROR) << rx_buffers[i]->str_code(rc);
+			LOGC(DDEV, ERROR) << rx_buffers[i]->str_status(timestamp);
 			return 0;
 		}
 	}
@@ -997,12 +785,12 @@ int uhd_device::writeSamples(std::vector<short *> &bufs, int len, bool *underrun
 
 	// No control packets
 	if (isControl) {
-		LOG(ERR) << "Control packets not supported";
+		LOGC(DDEV, ERROR) << "Control packets not supported";
 		return 0;
 	}
 
 	if (bufs.size() != chans) {
-		LOG(ALERT) << "Invalid channel combination " << bufs.size();
+		LOGC(DDEV, ALERT) << "Invalid channel combination " << bufs.size();
 		return -1;
 	}
 
@@ -1011,14 +799,14 @@ int uhd_device::writeSamples(std::vector<short *> &bufs, int len, bool *underrun
 		drop_cnt++;
 
 		if (drop_cnt == 1) {
-			LOG(DEBUG) << "Aligning transmitter: stop burst";
+			LOGC(DDEV, DEBUG) << "Aligning transmitter: stop burst";
 			*underrun = true;
 			metadata.end_of_burst = true;
 		} else if (drop_cnt < 30) {
-			LOG(DEBUG) << "Aligning transmitter: packet advance";
+			LOGC(DDEV, DEBUG) << "Aligning transmitter: packet advance";
 			return len;
 		} else {
-			LOG(DEBUG) << "Aligning transmitter: start burst";
+			LOGC(DDEV, DEBUG) << "Aligning transmitter: start burst";
 			metadata.start_of_burst = true;
 			aligned = true;
 			drop_cnt = 0;
@@ -1030,7 +818,7 @@ int uhd_device::writeSamples(std::vector<short *> &bufs, int len, bool *underrun
 	thread_enable_cancel(true);
 
 	if (num_smpls != (unsigned) len) {
-		LOG(ALERT) << "UHD: Device send timed out";
+		LOGC(DDEV, ALERT) << "UHD: Device send timed out";
 	}
 
 	return num_smpls;
@@ -1066,7 +854,7 @@ uhd::tune_request_t uhd_device::select_freq(double freq, size_t chan, bool tx)
 
 		return uhd::tune_request_t(freq, lo_offset);
 	} else if ((dev_type != B210) || (chans > 2) || (chan > 1)) {
-		LOG(ALERT) << chans << " channels unsupported";
+		LOGC(DDEV, ALERT) << chans << " channels unsupported";
 		return treq;
 	}
 
@@ -1082,7 +870,7 @@ uhd::tune_request_t uhd_device::select_freq(double freq, size_t chan, bool tx)
 	/* Find center frequency between channels */
 	rf_spread = fabs(freqs[!chan] - freq);
 	if (rf_spread > dev_param_map.at(dev_key(B210, tx_sps, rx_sps)).mcr) {
-		LOG(ALERT) << rf_spread << "Hz tuning spread not supported\n";
+		LOGC(DDEV, ALERT) << rf_spread << "Hz tuning spread not supported\n";
 		return treq;
 	}
 
@@ -1108,7 +896,7 @@ bool uhd_device::set_freq(double freq, size_t chan, bool tx)
 		tres = usrp_dev->set_rx_freq(treq, chan);
 		rx_freqs[chan] = usrp_dev->get_rx_freq(chan);
 	}
-	LOG(INFO) << "\n" << tres.to_pp_string() << std::endl;
+	LOGC(DDEV, INFO) << "\n" << tres.to_pp_string() << std::endl;
 
 	if ((chans == 1) || ((chans == 2) && dev_type == UMTRX))
 		return true;
@@ -1128,7 +916,7 @@ bool uhd_device::set_freq(double freq, size_t chan, bool tx)
 			rx_freqs[!chan] = usrp_dev->get_rx_freq(!chan);
 
 		}
-		LOG(INFO) << "\n" << tres.to_pp_string() << std::endl;
+		LOGC(DDEV, INFO) << "\n" << tres.to_pp_string() << std::endl;
 	}
 
 	return true;
@@ -1137,7 +925,7 @@ bool uhd_device::set_freq(double freq, size_t chan, bool tx)
 bool uhd_device::setTxFreq(double wFreq, size_t chan)
 {
 	if (chan >= tx_freqs.size()) {
-		LOG(ALERT) << "Requested non-existent channel " << chan;
+		LOGC(DDEV, ALERT) << "Requested non-existent channel " << chan;
 		return false;
 	}
 	ScopedLock lock(tune_lock);
@@ -1148,7 +936,7 @@ bool uhd_device::setTxFreq(double wFreq, size_t chan)
 bool uhd_device::setRxFreq(double wFreq, size_t chan)
 {
 	if (chan >= rx_freqs.size()) {
-		LOG(ALERT) << "Requested non-existent channel " << chan;
+		LOGC(DDEV, ALERT) << "Requested non-existent channel " << chan;
 		return false;
 	}
 	ScopedLock lock(tune_lock);
@@ -1159,7 +947,7 @@ bool uhd_device::setRxFreq(double wFreq, size_t chan)
 double uhd_device::getTxFreq(size_t chan)
 {
 	if (chan >= tx_freqs.size()) {
-		LOG(ALERT) << "Requested non-existent channel " << chan;
+		LOGC(DDEV, ALERT) << "Requested non-existent channel " << chan;
 		return 0.0;
 	}
 
@@ -1169,7 +957,7 @@ double uhd_device::getTxFreq(size_t chan)
 double uhd_device::getRxFreq(size_t chan)
 {
 	if (chan >= rx_freqs.size()) {
-		LOG(ALERT) << "Requested non-existent channel " << chan;
+		LOGC(DDEV, ALERT) << "Requested non-existent channel " << chan;
 		return 0.0;
 	}
 
@@ -1180,23 +968,23 @@ bool uhd_device::setRxAntenna(const std::string &ant, size_t chan)
 {
 	std::vector<std::string> avail;
 	if (chan >= rx_paths.size()) {
-		LOG(ALERT) << "Requested non-existent channel " << chan;
+		LOGC(DDEV, ALERT) << "Requested non-existent channel " << chan;
 		return false;
 	}
 
 	avail = usrp_dev->get_rx_antennas(chan);
 	if (std::find(avail.begin(), avail.end(), ant) == avail.end()) {
-		LOG(ALERT) << "Requested non-existent Rx antenna " << ant << " on channel " << chan;
-		LOG(INFO) << "Available Rx antennas: ";
+		LOGC(DDEV, ALERT) << "Requested non-existent Rx antenna " << ant << " on channel " << chan;
+		LOGC(DDEV, INFO) << "Available Rx antennas: ";
 		for (std::vector<std::string>::const_iterator i = avail.begin(); i != avail.end(); ++i)
-			LOG(INFO) << "- '" << *i << "'";
+			LOGC(DDEV, INFO) << "- '" << *i << "'";
 		return false;
 	}
 	usrp_dev->set_rx_antenna(ant, chan);
 	rx_paths[chan] = usrp_dev->get_rx_antenna(chan);
 
 	if (ant != rx_paths[chan]) {
-		LOG(ALERT) << "Failed setting antenna " << ant << " on channel " << chan << ", got instead " << rx_paths[chan];
+		LOGC(DDEV, ALERT) << "Failed setting antenna " << ant << " on channel " << chan << ", got instead " << rx_paths[chan];
 		return false;
 	}
 
@@ -1206,7 +994,7 @@ bool uhd_device::setRxAntenna(const std::string &ant, size_t chan)
 std::string uhd_device::getRxAntenna(size_t chan)
 {
 	if (chan >= rx_paths.size()) {
-		LOG(ALERT) << "Requested non-existent channel " << chan;
+		LOGC(DDEV, ALERT) << "Requested non-existent channel " << chan;
 		return "";
 	}
 	return usrp_dev->get_rx_antenna(chan);
@@ -1216,23 +1004,23 @@ bool uhd_device::setTxAntenna(const std::string &ant, size_t chan)
 {
 	std::vector<std::string> avail;
 	if (chan >= tx_paths.size()) {
-		LOG(ALERT) << "Requested non-existent channel " << chan;
+		LOGC(DDEV, ALERT) << "Requested non-existent channel " << chan;
 		return false;
 	}
 
 	avail = usrp_dev->get_tx_antennas(chan);
 	if (std::find(avail.begin(), avail.end(), ant) == avail.end()) {
-		LOG(ALERT) << "Requested non-existent Tx antenna " << ant << " on channel " << chan;
-		LOG(INFO) << "Available Tx antennas: ";
+		LOGC(DDEV, ALERT) << "Requested non-existent Tx antenna " << ant << " on channel " << chan;
+		LOGC(DDEV, INFO) << "Available Tx antennas: ";
 		for (std::vector<std::string>::const_iterator i = avail.begin(); i != avail.end(); ++i)
-			LOG(INFO) << "- '" << *i << "'";
+			LOGC(DDEV, INFO) << "- '" << *i << "'";
 		return false;
 	}
 	usrp_dev->set_tx_antenna(ant, chan);
 	tx_paths[chan] = usrp_dev->get_tx_antenna(chan);
 
 	if (ant != tx_paths[chan]) {
-		LOG(ALERT) << "Failed setting antenna " << ant << " on channel " << chan << ", got instead " << tx_paths[chan];
+		LOGC(DDEV, ALERT) << "Failed setting antenna " << ant << " on channel " << chan << ", got instead " << tx_paths[chan];
 		return false;
 	}
 
@@ -1242,7 +1030,7 @@ bool uhd_device::setTxAntenna(const std::string &ant, size_t chan)
 std::string uhd_device::getTxAntenna(size_t chan)
 {
 	if (chan >= tx_paths.size()) {
-		LOG(ALERT) << "Requested non-existent channel " << chan;
+		LOGC(DDEV, ALERT) << "Requested non-existent channel " << chan;
 		return "";
 	}
 	return usrp_dev->get_tx_antenna(chan);
@@ -1310,7 +1098,7 @@ bool uhd_device::recv_async_msg()
 
 		if ((md.event_code != uhd::async_metadata_t::EVENT_CODE_UNDERFLOW) &&
 		    (md.event_code != uhd::async_metadata_t::EVENT_CODE_TIME_ERROR)) {
-			LOG(ERR) << str_code(md);
+			LOGC(DDEV, ERROR) << str_code(md);
 		}
 	}
 
@@ -1386,166 +1174,6 @@ std::string uhd_device::str_code(uhd::async_metadata_t metadata)
 	return ost.str();
 }
 
-smpl_buf::smpl_buf(size_t len, double rate)
-	: buf_len(len), clk_rt(rate),
-	  time_start(0), time_end(0), data_start(0), data_end(0)
-{
-	data = new uint32_t[len];
-}
-
-smpl_buf::~smpl_buf()
-{
-	delete[] data;
-}
-
-ssize_t smpl_buf::avail_smpls(TIMESTAMP timestamp) const
-{
-	if (timestamp < time_start)
-		return ERROR_TIMESTAMP;
-	else if (timestamp >= time_end)
-		return 0;
-	else
-		return time_end - timestamp;
-}
-
-ssize_t smpl_buf::avail_smpls(uhd::time_spec_t timespec) const
-{
-	return avail_smpls(timespec.to_ticks(clk_rt));
-}
-
-ssize_t smpl_buf::read(void *buf, size_t len, TIMESTAMP timestamp)
-{
-	int type_sz = 2 * sizeof(short);
-
-	// Check for valid read
-	if (timestamp < time_start)
-		return ERROR_TIMESTAMP;
-	if (timestamp >= time_end)
-		return 0;
-	if (len >= buf_len)
-		return ERROR_READ;
-
-	// How many samples should be copied
-	size_t num_smpls = time_end - timestamp;
-	if (num_smpls > len)
-		num_smpls = len;
-
-	// Starting index
-	size_t read_start = (data_start + (timestamp - time_start)) % buf_len;
-
-	// Read it
-	if (read_start + num_smpls < buf_len) {
-		size_t numBytes = len * type_sz;
-		memcpy(buf, data + read_start, numBytes);
-	} else {
-		size_t first_cp = (buf_len - read_start) * type_sz;
-		size_t second_cp = len * type_sz - first_cp;
-
-		memcpy(buf, data + read_start, first_cp);
-		memcpy((char*) buf + first_cp, data, second_cp);
-	}
-
-	data_start = (read_start + len) % buf_len;
-	time_start = timestamp + len;
-
-	if (time_start > time_end)
-		return ERROR_READ;
-	else
-		return num_smpls;
-}
-
-ssize_t smpl_buf::read(void *buf, size_t len, uhd::time_spec_t ts)
-{
-	return read(buf, len, ts.to_ticks(clk_rt));
-}
-
-ssize_t smpl_buf::write(void *buf, size_t len, TIMESTAMP timestamp)
-{
-	int type_sz = 2 * sizeof(short);
-
-	// Check for valid write
-	if ((len == 0) || (len >= buf_len))
-		return ERROR_WRITE;
-	if ((timestamp + len) <= time_end)
-		return ERROR_TIMESTAMP;
-
-	if (timestamp < time_end) {
-		LOG(ERR) << "Overwriting old buffer data: timestamp="<<timestamp<<" time_end="<<time_end;
-		uhd::time_spec_t ts = uhd::time_spec_t::from_ticks(timestamp, clk_rt);
-		LOG(DEBUG) << "Requested timestamp = " << timestamp << " (real_sec=" << std::fixed << ts.get_real_secs() << " = " << ts.to_ticks(clk_rt) << ") rate=" << clk_rt;
-		// Do not return error here, because it's a rounding error and is not fatal
-	}
-	if (timestamp > time_end && time_end != 0) {
-		LOG(ERR) << "Skipping buffer data: timestamp="<<timestamp<<" time_end="<<time_end;
-		uhd::time_spec_t ts = uhd::time_spec_t::from_ticks(timestamp, clk_rt);
-		LOG(DEBUG) << "Requested timestamp = " << timestamp << " (real_sec=" << std::fixed << ts.get_real_secs() << " = " << ts.to_ticks(clk_rt) << ") rate=" << clk_rt;
-		// Do not return error here, because it's a rounding error and is not fatal
-	}
-
-	// Starting index
-	size_t write_start = (data_start + (timestamp - time_start)) % buf_len;
-
-	// Write it
-	if ((write_start + len) < buf_len) {
-		size_t numBytes = len * type_sz;
-		memcpy(data + write_start, buf, numBytes);
-	} else {
-		size_t first_cp = (buf_len - write_start) * type_sz;
-		size_t second_cp = len * type_sz - first_cp;
-
-		memcpy(data + write_start, buf, first_cp);
-		memcpy(data, (char*) buf + first_cp, second_cp);
-	}
-
-	data_end = (write_start + len) % buf_len;
-	time_end = timestamp + len;
-
-	if (!data_start)
-		data_start = write_start;
-
-	if (((write_start + len) > buf_len) && (data_end > data_start))
-		return ERROR_OVERFLOW;
-	else if (time_end <= time_start)
-		return ERROR_WRITE;
-	else
-		return len;
-}
-
-ssize_t smpl_buf::write(void *buf, size_t len, uhd::time_spec_t ts)
-{
-	return write(buf, len, ts.to_ticks(clk_rt));
-}
-
-std::string smpl_buf::str_status(size_t ts) const
-{
-	std::ostringstream ost("Sample buffer: ");
-
-	ost << "timestamp = " << ts;
-	ost << ", length = " << buf_len;
-	ost << ", time_start = " << time_start;
-	ost << ", time_end = " << time_end;
-	ost << ", data_start = " << data_start;
-	ost << ", data_end = " << data_end;
-
-	return ost.str();
-}
-
-std::string smpl_buf::str_code(ssize_t code)
-{
-	switch (code) {
-	case ERROR_TIMESTAMP:
-		return "Sample buffer: Requested timestamp is not valid";
-	case ERROR_READ:
-		return "Sample buffer: Read error";
-	case ERROR_WRITE:
-		return "Sample buffer: Write error";
-	case ERROR_OVERFLOW:
-		return "Sample buffer: Overrun";
-	default:
-		return "Sample buffer: Unknown error";
-	}
-}
-
 RadioDevice *RadioDevice::make(size_t tx_sps, size_t rx_sps,
 			       InterfaceType iface, size_t chans, double lo_offset,
 			       const std::vector<std::string>& tx_paths,
