@@ -22,16 +22,19 @@
 
 #include "Transceiver.h"
 #include "radioDevice.h"
+#include "Utils.h"
 
 #include <time.h>
 #include <signal.h>
 #include <stdlib.h>
 #include <unistd.h>
+#include <getopt.h>
 #include <sched.h>
 #include <vector>
 #include <string>
 #include <sstream>
 #include <iostream>
+#include <sys/signalfd.h>
 
 #include <GSMCommon.h>
 #include <Logger.h>
@@ -55,6 +58,7 @@ extern "C" {
 #include "convert.h"
 #include "trx_vty.h"
 #include "debug.h"
+#include "osmo_signal.h"
 }
 
 #define DEFAULT_CONFIG_FILE	"osmo-trx.cfg"
@@ -63,6 +67,7 @@ extern "C" {
 
 static char* config_file = (char*)DEFAULT_CONFIG_FILE;
 
+struct osmo_fd signal_ofd;
 volatile bool gshutdown = false;
 
 static void *tall_trx_ctx;
@@ -112,6 +117,20 @@ RadioInterface *makeRadioInterface(struct trx_ctx *trx,
 	return radio;
 }
 
+/* Callback function to be called every time we receive a signal from TRANSC */
+static int transc_sig_cb(unsigned int subsys, unsigned int signal,
+		     void *handler_data, void *signal_data)
+{
+	switch (signal) {
+	case S_TRANSC_STOP_REQUIRED:
+		gshutdown = true;
+                break;
+	default:
+                break;
+	}
+	return 0;
+}
+
 /* Create transceiver core
  *     The multi-threaded modem core operates at multiples of the GSM rate of
  *     270.8333 ksps and consists of GSM specific modulation, demodulation,
@@ -127,11 +146,13 @@ int makeTransceiver(struct trx_ctx *trx, RadioInterface *radio)
 			      trx->cfg.rx_sps, trx->cfg.num_chans, GSM::Time(3,0),
 			      radio, trx->cfg.rssi_offset);
 	if (!transceiver->init(trx->cfg.filler, trx->cfg.rtsc,
-		       trx->cfg.rach_delay, trx->cfg.egprs)) {
+		       trx->cfg.rach_delay, trx->cfg.egprs, trx->cfg.ext_rach)) {
 		LOG(ALERT) << "Failed to initialize transceiver";
 		return -1;
 	}
 
+        transceiver->setSignalHandler(transc_sig_cb);
+
 	for (size_t i = 0; i < trx->cfg.num_chans; i++) {
 		fifo = radio->receiveFIFO(i);
 		if (fifo && transceiver->receiveFIFO(fifo, i))
@@ -145,6 +166,12 @@ int makeTransceiver(struct trx_ctx *trx, RadioInterface *radio)
 
 static void sig_handler(int signo)
 {
+
+	if (gshutdown)
+		/* We are in the middle of shutdown process, avoid any kind of extra
+		   action like printing */
+		return;
+
 	fprintf(stdout, "signal %d received\n", signo);
 	switch (signo) {
 	case SIGINT:
@@ -160,43 +187,68 @@ static void sig_handler(int signo)
 	case SIGUSR2:
 		talloc_report_full(tall_trx_ctx, stderr);
 		break;
+	case SIGHUP:
+		log_targets_reopen();
 	default:
 		break;
 	}
+
+}
+
+static int signalfd_callback(struct osmo_fd *ofd, unsigned int what)
+{
+	struct signalfd_siginfo fdsi;
+	ssize_t s;
+
+	s = read(ofd->fd, &fdsi, sizeof(struct signalfd_siginfo));
+	if (s < 0) {
+		LOG(FATAL) << "Failed to read from signalfd ("<< ofd->fd << "): " << errno;
+		gshutdown = true;
+		return 0;
+	}
+	sig_handler(fdsi.ssi_signo);
+	return 0;
 }
 
 static void setup_signal_handlers()
 {
-	/* Handle keyboard interrupt SIGINT */
-	signal(SIGINT, &sig_handler);
-	signal(SIGTERM, &sig_handler);
+	sigset_t set;
+	int sfd;
+
 	signal(SIGABRT, &sig_handler);
-	signal(SIGUSR1, &sig_handler);
-	signal(SIGUSR2, &sig_handler);
 	osmo_init_ignore_signals();
-}
 
-static std::vector<std::string> comma_delimited_to_vector(char* opt)
-{
-	std::string str = std::string(opt);
-	std::vector<std::string> result;
-	std::stringstream ss(str);
+	/* Other threads created by this thread (main) will inherit a copy of the
+	signal mask. */
+	sigemptyset(&set);
+	sigaddset(&set, SIGINT);
+	sigaddset(&set, SIGTERM);
+	sigaddset(&set, SIGUSR1);
+	sigaddset(&set, SIGUSR2);
+	sigaddset(&set, SIGHUP);
+	if (pthread_sigmask(SIG_BLOCK, &set, NULL)) {
+		fprintf(stderr, "pthread_sigmask() failed.\n");
+		exit(EXIT_FAILURE);
+	}
 
-	while( ss.good() )
-	{
-	    std::string substr;
-	    getline(ss, substr, ',');
-	    result.push_back(substr);
+	if ((sfd = signalfd(-1, &set, 0)) == -1) {
+		fprintf(stderr, "signalfd() failed (%d).\n", errno);
+		exit(EXIT_FAILURE);
+	}
+
+	osmo_fd_setup(&signal_ofd, sfd, BSC_FD_READ, signalfd_callback, NULL, 0);
+	if (osmo_fd_register(&signal_ofd) < 0) {
+		fprintf(stderr, "osmo_fd_register() failed.\n");
+		exit(EXIT_FAILURE);
 	}
-	return result;
 }
 
 static void print_help()
 {
 	fprintf(stdout, "Options:\n"
-		"  -h    This text\n"
-		"  -C    Filename The config file to use\n"
-		"  -V    Print the version of OsmoTRX\n"
+		"  -h, --help      This text\n"
+		"  -C, --config    Filename The config file to use\n"
+		"  -V, --version   Print the version of OsmoTRX\n"
 		);
 }
 
@@ -213,8 +265,15 @@ static void handle_options(int argc, char **argv, struct trx_ctx* trx)
 	unsigned int i;
 	std::vector<std::string> rx_paths, tx_paths;
 	bool rx_paths_set = false, tx_paths_set = false;
+	static struct option long_options[] = {
+		{"help", 0, 0, 'h'},
+		{"config", 1, 0, 'C'},
+		{"version", 0, 0, 'V'},
+		{NULL, 0, 0, 0}
+	};
 
-	while ((option = getopt(argc, argv, "ha:l:i:j:p:c:dmxgfo:s:b:r:A:R:Set:y:z:C:V")) != -1) {
+	while ((option = getopt_long(argc, argv, "ha:l:i:j:p:c:dmxgfo:s:b:r:A:R:Set:y:z:C:V", long_options,
+		NULL)) != -1) {
 		switch (option) {
 		case 'h':
 			print_help();
@@ -352,14 +411,14 @@ bad_config:
 
 int trx_validate_config(struct trx_ctx *trx)
 {
-	if (trx->cfg.multi_arfcn && trx->cfg.num_chans > 5) {
+	if (trx->cfg.multi_arfcn && trx->cfg.num_chans > TRX_MCHAN_MAX) {
 		LOG(ERROR) << "Unsupported number of channels";
 		return -1;
 	}
 
 	/* Force 4 SPS for EDGE or multi-ARFCN configurations */
 	if ((trx->cfg.egprs || trx->cfg.multi_arfcn) &&
-	    (trx->cfg.tx_sps!=4 || trx->cfg.tx_sps!=4)) {
+	    (trx->cfg.tx_sps!=4 || trx->cfg.rx_sps!=4)) {
 		LOG(ERROR) << "EDGE and Multi-Carrier options require 4 tx and rx sps. Check you config.";
 		return -1;
 	}
@@ -373,7 +432,7 @@ static int set_sched_rr(unsigned int prio)
 	int rc;
 	memset(&param, 0, sizeof(param));
 	param.sched_priority = prio;
-	printf("Setting SCHED_RR priority(%d)\n", param.sched_priority);
+	LOG(INFO) << "Setting SCHED_RR priority " << param.sched_priority;
 	rc = sched_setscheduler(getpid(), SCHED_RR, &param);
 	if (rc != 0) {
 		LOG(ERROR) << "Config: Setting SCHED_RR failed";
@@ -397,6 +456,7 @@ static void print_config(struct trx_ctx *trx)
 	ost << "   Tx Samples-per-Symbol... " << trx->cfg.tx_sps << std::endl;
 	ost << "   Rx Samples-per-Symbol... " << trx->cfg.rx_sps << std::endl;
 	ost << "   EDGE support............ " << trx->cfg.egprs << std::endl;
+	ost << "   Extended RACH support... " << trx->cfg.ext_rach << std::endl;
 	ost << "   Reference............... " << trx->cfg.clock_ref << std::endl;
 	ost << "   C0 Filler Table......... " << trx->cfg.filler << std::endl;
 	ost << "   Multi-Carrier........... " << trx->cfg.multi_arfcn << std::endl;
@@ -416,12 +476,12 @@ static void print_config(struct trx_ctx *trx)
 	}
 	ost << std::endl;
 
-	std::cout << ost << std::endl;
+	LOG(INFO) << ost << std::endl;
 }
 
 static void trx_stop()
 {
-	std::cout << "Shutting down transceiver..." << std::endl;
+	LOG(NOTICE) << "Shutting down transceiver..." << std::endl;
 
 	delete transceiver;
 	delete radio;
@@ -464,7 +524,7 @@ static int trx_start(struct trx_ctx *trx)
 		goto shutdown;
 
 	chans = transceiver->numChans();
-	std::cout << "-- Transceiver active with "
+	LOG(NOTICE) << "-- Transceiver active with "
 		  << chans << " channel(s)" << std::endl;
 
 	return 0;
@@ -539,7 +599,7 @@ int main(int argc, char *argv[])
 
 	g_ctrlh = ctrl_interface_setup(NULL, OSMO_CTRL_PORT_TRX, NULL);
 	if (!g_ctrlh) {
-		fprintf(stderr, "Failed to create CTRL interface.\n");
+		LOG(ERROR) << "Failed to create CTRL interface.\n";
 		exit(1);
 	}
 
@@ -576,5 +636,7 @@ int main(int argc, char *argv[])
 
 	trx_stop();
 
+	osmo_fd_unregister(&signal_ofd);
+	osmo_fd_close(&signal_ofd);
 	return 0;
 }
